# Autogenerated using:
# - Template: 'src/cubnm/sim/simgroup.py.mako'
# - Model definition: 'codegen/recipes/wc.yaml'
# Do not modify this autogenerated code. Instead modify template
# or model definition.
from .base import SimGroup
import sys
import logging
import numpy as np

logging.basicConfig(stream=sys.stdout, level=logging.WARNING)
logger = logging.getLogger(__name__)

class WCSimGroup(SimGroup):
    model_name = "WC"
    global_param_names = ['G']
    regional_param_names = ['c_EE', 'c_IE', 'c_EI', 'c_II', 'P_E', 'P_I', 'sigma_E', 'sigma_I']
    state_names = ['E', 'I']
    sel_state_var = "E"
    n_noise = 2
    default_params = {
        "G": None,  # required
        "c_EE": 16,
        "c_IE": 12,
        "c_EI": 15,
        "c_II": 3,
        "P_E": 0.5,
        "P_I": 0.0,
        "sigma_E": 5e-05,
        "sigma_I": 5e-05,
    }

    def __init__(self, *args, **kwargs):
        r"""
        Group of Wilson-Cowan simulations that are executed in parallel.

        Parameters
        ---------
        *args, **kwargs:
            see :class:`cubnm.sim.SimGroup` for details

        Attributes
        ----------
        param_lists: :obj:`dict` of :obj:`np.ndarray`
            dictionary of parameter lists, including
                - ``'G'``: global coupling strength. Shape: (N_SIMS,)
                - ``'c_EE'``: local E to E coupling. Shape: (N_SIMS, nodes)
                - ``'c_IE'``: local I to E coupling. Shape: (N_SIMS, nodes)
                - ``'c_EI'``: local E to I coupling. Shape: (N_SIMS, nodes)
                - ``'c_II'``: local I to I coupling. Shape: (N_SIMS, nodes)
                - ``'P_E'``: excitatory background drive. Shape: (N_SIMS, nodes)
                - ``'P_I'``: inhibitory background drive. Shape: (N_SIMS, nodes)
                - ``'sigma_E'``: excitatory noise strength. Shape: (N_SIMS, nodes)
                - ``'sigma_I'``: inhibitory noise strength. Shape: (N_SIMS, nodes)
                - ``'v'``: conduction velocity. Shape: (N_SIMS,)

        Equations
        ---------
        .. math::

            \begin{gather}
            \tau^E\dot{E_i} = -E + (1 - E) S^E[c^{EE} E_i - c^{IE}I_i + G \sum_j{C_{ij} E_j} + P_i^E] + \sigma_i^E\epsilon_i^E \\
            \tau^I\dot{I_i} = -I + (1 - I) S^I[c^{EI} E_i - c^{II}I_i + P_i^I] + \sigma_i^I\epsilon_i^I, \text{where:} \\
            S^E(x) = \frac{1}{1 + e^{-a^E(x - \mu^E)}} \\
            S^I(x) = \frac{1}{1 + e^{-a^I(x - \mu^I)}} \\
            \end{gather}

        References
        ----------        
        * Wilson et al. 1972 Biophysical Journal (10.1016/S0006-3495(72)86068-5)
        * Papadopoulos et al. 2020 PLOS Computational Biology (10.1371/journal.pcbi.1008144)
        """
        # parent init must be called after setting
        # model-specific attributes because
        # it sets .last_config which may include some
        # model-specific attributes
        super().__init__(*args, **kwargs)

    @property
    def labels(self):
        """
        Labels of parameters and state variables
        to use in plots and reports
        """
        labels = super().labels
        labels.update({
            'G': r'G',
            'E': r'E',
            'I': r'I',
            'c_EE': r'$c_{EE}$',
            'c_IE': r'$c_{IE}$',
            'c_EI': r'$c_{EI}$',
            'c_II': r'$c_{II}$',
            'P_E': r'$P_E$',
            'P_I': r'$P_I$',
            'sigma_E': r'$\sigma_E$',
            'sigma_I': r'$\sigma_I$',
        })
        return labels



# Autogenerated using:
# - Template: 'src/cubnm/sim/simgroup.py.mako'
# - Model definition: 'codegen/recipes/jr.yaml'
# Do not modify this autogenerated code. Instead modify template
# or model definition.
from .base import SimGroup
import sys
import logging
import numpy as np

logging.basicConfig(stream=sys.stdout, level=logging.WARNING)
logger = logging.getLogger(__name__)

class JRSimGroup(SimGroup):
    model_name = "JR"
    global_param_names = ['G']
    regional_param_names = ['J', 'a_1', 'a_2', 'a_3', 'a_4', 'sigma']
    state_names = ['y0', 'y1', 'y2', 'y3', 'y4', 'y5', 'y1_y2', 's_y1_y2']
    sel_state_var = "y1_y2"
    n_noise = 1
    default_params = {
        "G": None,  # required
        "J": 135.0,
        "a_1": 1.0,
        "a_2": 0.8,
        "a_3": 0.25,
        "a_4": 0.25,
        "sigma": 0.0001,
    }

    def __init__(self, *args, **kwargs):
        r"""
        Group of Jansen-Rit simulations that are executed in parallel.

        Parameters
        ---------
        *args, **kwargs:
            see :class:`cubnm.sim.SimGroup` for details

        Attributes
        ----------
        param_lists: :obj:`dict` of :obj:`np.ndarray`
            dictionary of parameter lists, including
                - ``'G'``: global coupling strength. Shape: (N_SIMS,)
                - ``'J'``: average number of synapses. Shape: (N_SIMS, nodes)
                - ``'a_1'``: average probability of synapses in feedback excitatory loop. Shape: (N_SIMS, nodes)
                - ``'a_2'``: average probability of synapses in slow feedback excitatory loop. Shape: (N_SIMS, nodes)
                - ``'a_3'``: average probability of synapses in feedback inhibitory loop. Shape: (N_SIMS, nodes)
                - ``'a_4'``: average probability of synapses in slow feedback inhibitory loop. Shape: (N_SIMS, nodes)
                - ``'sigma'``: local noise sigma. Shape: (N_SIMS, nodes)
                - ``'v'``: conduction velocity. Shape: (N_SIMS,)

        Equations
        ---------
        .. math::

            \begin{gather}
            \dot{y_{0_{i}}} = y_{3_{i}} \\
            \dot{y_{1_{i}}} = y_{4_{i}} \\
            \dot{y_{2_{i}}} = y_{5_{i}} \\
            \dot{y_{3_{i}}} = A a [\frac{2 \nu_{max}}{1+e^{r(v_0 - (y_{2_{i}} - y_{1_{i}}))}}]-2 a y_{3_{i}}-a^2 y_{0_{i}} + \sigma_i\epsilon_i \\
            \dot{y_{4_{i}}} = A a [\mu+a_{2_{i}} J_i \frac{2\nu_{max}}{1+e^{r(v_0 - (a_{1_{i}} J y_{1_{i}}))}}+ \sum_j{C_{ij}x_j} \\
            \dot{y_{5_{i}}} = B b (a_{4_{i}} J_i [\frac{2\nu_{max}}{1 + e^{r(v_0 - (a_{3_{i}} J y_{0_{i}})}}]) - 2by_{5_{i}} - b^2 y_{2_{i}} \\
            z_i = y_{1_{i}} - y_{2_{i}} \\
            x_i = c_{min} + \frac{c_{max}-c_{min}}{1 + e^{r(c_{mid} -z_i)}} \\
            \end{gather}

        References
        ----------        
        * Jansen and Rit 1995 Biological Cybernetics (10.1007/BF00199471)
        * Ritter et al. 2013 Brain Connectivity (10.1089/brain.2012.0120)
        """
        # parent init must be called after setting
        # model-specific attributes because
        # it sets .last_config which may include some
        # model-specific attributes
        super().__init__(*args, **kwargs)



# Autogenerated using:
# - Template: 'src/cubnm/sim/simgroup.py.mako'
# - Model definition: 'codegen/recipes/rwwex.yaml'
# Do not modify this autogenerated code. Instead modify template
# or model definition.
from .base import SimGroup
import sys
import logging
import numpy as np

logging.basicConfig(stream=sys.stdout, level=logging.WARNING)
logger = logging.getLogger(__name__)

class rWWExSimGroup(SimGroup):
    model_name = "rWWEx"
    global_param_names = ['G']
    regional_param_names = ['w', 'I0', 'sigma']
    state_names = ['x', 'r', 'S']
    sel_state_var = "r"
    n_noise = 1
    default_params = {
        "G": None,  # required
        "w": 0.9,
        "I0": 0.3,
        "sigma": 0.001,
    }

    def __init__(self, *args, **kwargs):
        r"""
        Group of reduced Wong Wang (excitatory) simulations that are executed in parallel.

        Parameters
        ---------
        *args, **kwargs:
            see :class:`cubnm.sim.SimGroup` for details

        Attributes
        ----------
        param_lists: :obj:`dict` of :obj:`np.ndarray`
            dictionary of parameter lists, including
                - ``'G'``: global coupling strength. Shape: (N_SIMS,)
                - ``'w'``: local excitatory recurrence. Shape: (N_SIMS, nodes)
                - ``'I0'``: external input current. Shape: (N_SIMS, nodes)
                - ``'sigma'``: noise amplitude. Shape: (N_SIMS, nodes)
                - ``'v'``: conduction velocity. Shape: (N_SIMS,)

        Equations
        ---------
        .. math::

            \begin{gather}
            x_i = w_i J^N S_i + G J^N \sum_j C_{ij} S_j + I_{0_{i}} \\
            r_i = \frac{a x_i - b}{1 - e^{-d(a x_i - b)}} \\
            \dot{S_i} = -\frac{S_i}{\tau_s} + (1 - S_i) \gamma r_i + {\sigma}_i \epsilon_i \\
            \end{gather}

        References
        ----------        
        * Deco et al. 2013 Journal of Neuroscience (10.1523/JNEUROSCI.1091-13.2013)
        """
        # parent init must be called after setting
        # model-specific attributes because
        # it sets .last_config which may include some
        # model-specific attributes
        super().__init__(*args, **kwargs)



"""
CLI parser

Note: This is in a separate file so that sphinx doesn't require pandas and cubnm
for autogenerating the CLI docs
"""
import argparse
import os
import ast
import sys
if sys.version_info >= (3, 9):
    from importlib.resources import files
else:
    from importlib_resources import files

def parse_params(param_str):
    """
    Parses input string of parameters and their
    ranges into a dictionary recognizable by GridSearch
    and BNMProblem

    Parameters
    ----------
    param_str : :obj:`str`
        a string with the format "<param_name>=<start>[:<end>[:<step>]],..."

    Returns
    -------
    :obj:`dict` with the format {<param_name>:(start[,end[,step]]),...}
    """
    params = {}
    for item in param_str.split(','):
        key, value = item.split('=')
        if ':' in value:
            params[key] = list(map(float, value.split(':')))
            assert len(params[key]) <= 3
            if len(params[key]) == 3:
                params[key][2] = int(params[key][2])
            params[key] = tuple(params[key])
        else:
            params[key] = float(value)
    return params

def parse_maps_coef_range(value):
    """
    Parses input string of map coefficient ranges

    Parameters
    ----------
    value : :obj:`str`
        - "auto"
        - "<start>:<end>": same range for all maps
        - "<start1>:<end1>,<start2>:<end2>,...": different range for each map

    Returns
    -------
    "auto", :obj:`tuple` or :obj:`list` of :obj:`tuple`
    """
    if value == 'auto':
        return value
    try:
        maps_coef_range = [tuple(map(float, item.split(':'))) for item in value.split(',')]
    except ValueError:
        raise argparse.ArgumentTypeError("Value should be 'auto' or a list of tuples")
    if len(maps_coef_range) == 1:
        return maps_coef_range[0]
    else:
        return maps_coef_range

def parse_grid_shape(value):
    """
    Parses input string of grid shape

    Parameters
    ----------
    value : :obj:`str`
        - "<int>": same number of points for each parameter
        - "<param_name=size>,...": different number of points for each parameter
    
    Returns
    -------
    :obj:`int` or :obj:`dict`
    """
    try:
        grid_shape = int(value)
    except ValueError:
        grid_shape = {}
        for item in value.split(','):
            key, value = item.split('=')
            grid_shape[key] = int(value)
    return grid_shape

def get_class_names(file_path):
    """
    Returns a list of class names in a python file
    without importing it

    Note: This is important for generating the documentation
    as in the docs environment we don't have the package
    installed (and don't want to have it installed as that
    would require a complicated setup of readthedocs)

    Parameters
    ----------
    file_path : :obj:`str`
        path to the python file
    
    Returns
    -------
    :obj:`list` of :obj:`str`
    """
    with open(file_path, "r", encoding="utf-8") as file:
        tree = ast.parse(file.read(), filename=file_path)
    class_names = [node.name for node in ast.walk(tree) if isinstance(node, ast.ClassDef)]
    return class_names

def get_models():
    """
    Returns a list of available models

    Returns
    -------
    :obj:`list` of :obj:`str`
    """
    if os.environ.get("READTHEDOCS") == "True":
        init_py_path = os.path.join(
            os.environ["READTHEDOCS_REPOSITORY_PATH"], 
            "src", "cubnm", "sim", "__init__.py"
        )
    else:
        try:
            # when user runs the CLI
            init_py_path = os.path.abspath(
                files("cubnm").joinpath("sim", "__init__.py").as_posix()
            )
        except ModuleNotFoundError:
            # when running sphinx locally current directory is docs
            init_py_path = os.path.abspath(
                os.path.join("..", "src", "cubnm", "sim", "__init__.py")
            )
    
    # parse __all__ from __init__.py
    with open(init_py_path, "r", encoding="utf-8") as file:
        tree = ast.parse(file.read(), filename=init_py_path)
    
    model_names = []
    for node in ast.walk(tree):
        if isinstance(node, ast.Assign):
            for target in node.targets:
                if isinstance(target, ast.Name) and target.id == '__all__':
                    if isinstance(node.value, ast.List):
                        model_names = [
                            elt.s if hasattr(elt, 's') else elt.value 
                            for elt in node.value.elts 
                            if isinstance(elt, (ast.Str, ast.Constant))
                        ]
                    break
    
    # Filter to only keep SimGroup classes (excluding base SimGroup)
    model_names = [
        name.replace("SimGroup", "") 
        for name in model_names 
        if name.endswith("SimGroup") and name != "SimGroup"
    ]
    
    return model_names


def add_shared_arguments(parser):
    # simgroup arguments shared between all commands
    parser.add_argument('-m', '--model', type=str, required=True, 
                        choices=get_models(),
                        help='Model (required)')
    parser.add_argument('-p', '--params', type=parse_params, required=True, 
                        help='Parameters in custom format inside'
                        ' quotation marks (required). Format:'
                         ' "<param_name>=<start>[:<end>],..."'
                       )
    parser.add_argument('--duration', type=int, required=True, 
                        help='Duration of the simulation (required)')
    parser.add_argument('--TR', type=float, required=True, 
                        help='Repetition time (required)')
    parser.add_argument('--sc', type=str, required=True, 
                        help='Structural connectivity .txt file or "example" (required)')
    parser.add_argument('--sc_dist', type=str, 
                        help='Structural connectivity distribution .txt file or "example"')
    parser.add_argument('--dt', type=str, default='0.1', 
                        help='Neuronal model integration step (msec)')
    parser.add_argument('--bw_dt', type=str, default='1.0', 
                        help='Balloon-Windkessel model integration step (msec)')
    parser.add_argument('-d', '--out_dir', type=str, default="same", 
                        help='Output directory')
    parser.add_argument('--emp_fc_tril', type=str,
                    help='Functional connectivity lower triangle as space-separated .txt file')
    parser.add_argument('--emp_fcd_tril', type=str,
                    help='Functional connectivity dynamics lower triangle as space-separated .txt file')
    parser.add_argument('--emp_bold', type=str,
                    help='Cleaned and parcellated BOLD signal as space-separated .txt file or "example"'
                        ' BOLD signal should be in the shape (nodes, volumes).'
                        ' Motion outliers can either be excluded (not recommended as it disrupts'
                        ' the temporal structure) or replaced with zeros.'
                        ' If provided emp_fc_tril and emp_fcd_tril will be ignored.')
    parser.add_argument('--no_fc', action='store_true', 
                        help='Do not calculate simulated/empirical FC')
    parser.add_argument('--no_fcd', action='store_true', 
                        help='Do not calculate simulated/empirical FCD')
    parser.add_argument('--sim_seed', type=int, default=0, 
                        help='Simulation noise seed')
    parser.add_argument('--noise_segment_length', type=int, default=30, 
                        help='Noise segment length (seconds)')
    parser.add_argument('--states_ts', action='store_true',
                        help='Return model states timeseries')
    parser.add_argument('--states_sampling', type=float, 
                        help='Sampling rate of model state variables')
    parser.add_argument('--bw_params', type=str, default='friston2003', 
                        choices=['friston2003', 'heinzle2016-3T'],
                        help='Balloon Windkessel parameters')
    parser.add_argument('--bold_remove_s', type=int, default=30, 
                        help='Remove initial n seconds of the simulation from FC/FCD calculations'
                           ' and average of state variables')
    parser.add_argument('--window_size', type=int, default=30, 
                        help='FCD window size (s)')
    parser.add_argument('--window_step', type=int, default=5, 
                        help='FCD window step (s)')
    parser.add_argument('--exc_interhemispheric', action='store_true', 
                        help='Exclude interhemispheric connections in FC/FCD calculations')
    parser.add_argument('--fcd_keep_edges', action='store_true', 
                        help='Keep edge windows from FCD calculations')
    parser.add_argument('--gof_terms',
                        type=lambda s: s.split(','),
                        default=["fc_corr","fcd_ks"],
                        help='Goodness of fit terms (comma separated in quotation marks)'
                            ' e.g. "fc_corr,fcd_ks".'
                       )
    parser.add_argument('--force_cpu', action='store_true', 
                        help='Force CPU')
    parser.add_argument('--force_gpu', action='store_true', 
                        help='Force GPU')
    parser.add_argument('-v', '--sim_verbose', action='store_true',
                        help='Show simulation progress')
    parser.add_argument('--progress_interval', type=int, default=500, 
                        help='Interval of progress updates (in msec).'
                            ' Only used if sim_verbose is True'
                       )
    parser.add_argument('--no_print_args', action='store_true', 
                        help='Do not print command line arguments table')
    # BNMProblem arguments
    parser.add_argument('--het_params', nargs='+', default=[],
                        help='List of heterogeneous regional parameters (space separated)')
    parser.add_argument('--maps', type=str, 
                        help='Path to heterogeneity maps or "example"')
    parser.add_argument('--maps_coef_range', type=parse_maps_coef_range, 
                        default='auto',
                        help='Coefficient range for maps.'
                        ' Options: "auto", "min:max" (same for all maps),'
                        ' "min1:max1,min2:max2,..." (different for each map)'
                        )
    parser.add_argument('--node_grouping', type=str, 
                        default=None,
                        help='Path to node grouping array or special values: node, sym')
    parser.add_argument('--multiobj', action='store_true', 
                        help='Instead of combining the objectives into a single objective'
                        ' function (via summation) defines each objective separately.'
                        ' This must not be used with single-objective optimizers'
                        )


def get_parser():
    parser = argparse.ArgumentParser(
        description='cuBNM command line interface',
    )
    # add specific commands and their options
    subparsers = parser.add_subparsers(dest='cmd')

    # optimizer command
    parser_optimize = subparsers.add_parser('optimize', help='Optimize parameters using evolutionary algorithms')
    add_shared_arguments(parser_optimize)
    # Optimizer arguments
    parser_optimize.add_argument('-o', '--optimizer', type=str, required=True,
                                help='Optimizer type')
    parser_optimize.add_argument('--optimizer_seed', type=int, default=1, 
                                help='Optimizer random seed')
    parser_optimize.add_argument('--n_iter', type=int, default=80, 
                                help='Optimizer max iterations')
    parser_optimize.add_argument('--popsize', type=int, default=24, 
                                help='Optimizer population size')

    # grid command
    parser_grid = subparsers.add_parser('grid', help='Grid search')
    add_shared_arguments(parser_grid)
    parser_grid.add_argument('--grid_shape', type=parse_grid_shape, required=True,
                            help='Shape of the grid search: <int> (same number of'
                                'points for each parameter) or "<param_name=size>,..."'
                                ' (different number of points for each parameter)'
                            )
    
    # TODO: add model specific options (e.g. do_fic in rWW)
    # TODO: add optimizer specific options
    
    # TODO: add run command
    # parser_run = subparsers.add_parser('run', help='Run one simulation with given parameters')

    return parser


"""
Utility functions
"""
import subprocess
import json
import gc
import sys
import logging
import numpy as np
import scipy

from cubnm._setup_opts import gpu_model_flag
try:
    import cupy as cp
    from numba import cuda
    has_cupy = True
except ImportError:
    cp = None
    cuda = None
    has_cupy = False

logging.basicConfig(stream=sys.stdout, level=logging.WARNING)
logger = logging.getLogger(__name__)

if has_cupy:
    @cuda.jit
    def cdf_2d_2d_right(A_sorted, x_values, out):
        """
        CUDA kernel for calculating CDF of matching rows in 2D arrays.
        For each row i, performs searchsorted (right-sided) of 
        x_values[i] in A_sorted[i] divided by the length of A_sorted[i]
        resulting in empirical CDF of A_sorted[i] at x_values[i].
        Each row of A_sorted[i] must be sorted in ascending order.

        Parameters
        ----------
        A_sorted: :obj:`cp.ndarray`
            2D array of sorted values. Shape: (n_rows, m)
        x_values: :obj:`cp.ndarray`
            2D array of values to calculate CDF for. Shape: (n_rows, k)
        out: :obj:`cp.ndarray`
            2D array to store the CDF values. Shape: (n_rows, k)
        """
        # get current row and x_value index (col)
        row, col = cuda.grid(2)
        # get dimensions
        n_rows, m = A_sorted.shape
        _, k = x_values.shape
        # calculate CDF of A_sorted[row]
        # at x_values[row, col]
        if row < n_rows and col < k:
            x = x_values[row, col]
            left = 0
            right = m
            while left < right:
                mid = (left + right) // 2
                if x < A_sorted[row, mid]:
                    right = mid
                else:
                    left = mid + 1
            out[row, col] = left / m
else:
    def cdf_2d_2d_right(*args, **kwargs):
        raise RuntimeError(
            "CuPy is not available. Install `cubnm[cupy-cuda11x]` or `cubnm[cupy-cuda12x]`."
        )

def avail_gpus():
    """
    Get the number of available GPUs

    Returns
    -------
    :obj:`int`
        Number of available GPUs
    """
    gpu_counts = {'nvidia': 0, 'rocm': 0}
    # nvidia
    try:
        output = subprocess.check_output(['nvidia-smi', '-L'])
        output = output.decode('utf-8').strip()
        gpu_list = output.split('\n')
        gpu_counts['nvidia'] = len(gpu_list)
    except subprocess.CalledProcessError:
        gpu_counts['nvidia'] = 0
    except FileNotFoundError:
        gpu_counts['nvidia'] = 0
    # rocm
    try:
        output = subprocess.check_output(['rocm-smi', '-i', '--json'])
        output = output.decode('utf-8').strip()
        gpu_dict = json.loads(output)
        gpu_counts['rocm'] = len(gpu_dict)
    except subprocess.CalledProcessError:
        gpu_counts['rocm'] = 0
    except FileNotFoundError:
        gpu_counts['rocm'] = 0
    # warn if there is mismatch between the type of
    # gpu for which the code is complied vs available gpus
    if (
        ((gpu_model_flag == 'nvidia') and (gpu_counts['rocm'] > 0)) |
        ((gpu_model_flag == 'rocm') and (gpu_counts['nvidia'] > 0))
    ):
        logger.warning(
            f"Toolbox compiled for {gpu_model_flag} "
            f"but {(set(gpu_counts.keys()) - set([gpu_model_flag]))} "
            "GPUs are available. Reinstall for the correct GPU model."
        )
    # but only return the number of gpus for which
    # the code was compiled
    return gpu_counts[gpu_model_flag]

def is_jupyter():
    """
    Check if the current environment is a Jupyter notebook

    Returns
    -------
    :obj:`bool`
    """
    try:
        from IPython import get_ipython
    except ImportError:
        return False
    if hasattr(get_ipython(), 'config'):
        if 'IPKernelApp' in get_ipython().config:  
            return True
    return False

def get_bw_params(src):
    """
    Get Balloon-Windkessel model parameters

    Parameters
    ----------
    src: {'friston2003', 'heinzle2016-3T'}
        Source of the Balloon-Windkessel model parameters.
        - ``'friston2003'``: Friston et al. 2003
        - ``'heinzle2016-3T'``: Heinzle et al. 2016, 3T parameters

    Returns
    -------
    :obj:`dict`
        Balloon-Windkessel model parameters
    """
    if src == 'friston2003':
        rho = 0.34
        return {
            'k1': 7*rho,
            'k2': 2.0,
            'k3': 2*rho - 0.2,
        }
    elif src == 'heinzle2016-3T':
        return {
            'k1': 3.72,
            'k2': 0.527,
            'k3': 0.53,
        }
    else:
        raise ValueError(f'Unknown BW parameters source: {src}')

def calculate_fc(
    bold,
    exc_interhemispheric=False,
    return_tril=True,
):
    """
    Calculates functional connectivity matrix

    Parameters
    ---------
    bold: :obj:`np.ndarray`
        cleaned and parcellated empirical BOLD time series. Shape: (nodes, volumes)
        Motion outliers should either be excluded or replaced with zeros.
    exc_interhemispheric: :obj:`bool`
        exclude interhemispheric connections
    return_tril: :obj:`bool`
        return only the lower triangular part of the FCD matrix

    Returns
    -------
    :obj:`np.ndarray`
        FC dynamics matrix. 
        Shape: (nodes, nodes) or (n_node_pairs,)
        if ``return_tril`` is ``True``
    """
    # z-score non-zero volumes in each node
    outlier_vols = bold.sum(axis=0) == 0
    bold[:, ~outlier_vols] = scipy.stats.zscore(bold[:, ~outlier_vols], axis=1)
    # calculate FC
    fc = np.corrcoef(bold)
    if exc_interhemispheric:
        # set interhemispheric connections to NaN
        rh_idx = bold.shape[0] // 2
        fc[:rh_idx, rh_idx:] = np.NaN
        fc[rh_idx:, :rh_idx] = np.NaN
    if return_tril:
        fc = fc[np.tril_indices(fc.shape[0], -1)]
        # drop NaNs (interhemispheric connections)
        fc = fc[~np.isnan(fc)]
    return fc


def calculate_fcd(
    bold,
    window_size,
    window_step,
    drop_edges=True,
    outlier_threshold=0.5,
    exc_interhemispheric=False,
    return_tril=True,
    return_dfc=False,
):
    """
    Calculates functional connectivity dynamics matrix
    and dynamic functional connectivity matrices

    Parameters
    ---------
    bold: :obj:`np.ndarray`
        cleaned and parcellated empirical BOLD time series. Shape: (nodes, volumes)
        Motion outliers should either be excluded (not recommended as it disrupts
        the temporal structure) or replaced with zeros.
    window_size: :obj:`int`
        dynamic FC window size (in TR)
        Must be even. The actual window size is +1 (including center).
    window_step: :obj:`int`
        dynamic FC window step (in TR)
    drop_edges: :obj:`bool`
        drop edge windows which have less than window_size volumes
    outlier_threshold: :obj:`float`
        threshold for the proportion of motion outliers in a window
        that would lead to discarding the window
    exc_interhemispheric: :obj:`bool`
        exclude interhemispheric connections
    return_tril: :obj:`bool`
        return only the lower triangular part of the FCD matrix
    return_dfc: :obj:`bool`
        return dynamic FCs as well

    Returns
    -------
    :obj:`np.ndarray`
        FC dynamics matrix. 
        Shape: (n_windows, n_windows) or (n_window_pairs,)
        if ``return_tril`` is ``True``
    :obj:`np.ndarray`
        dynamic FCs. Shape: (nodes, nodes, n_windows)
        Returned only if ``return_dfc`` is ``True``
    """
    # TODO: make non-even window size work
    assert window_size % 2 == 0, "Window size must be even"
    # z-score non-zero volumes in each node
    outlier_vols = bold.sum(axis=0) == 0
    bold[:, ~outlier_vols] = scipy.stats.zscore(bold[:, ~outlier_vols], axis=1)
    # calculate dynamic FC
    nodes = bold.shape[0]
    n_vols = bold.shape[1]
    window_fc_trils = []
    window_fcs = []
    if drop_edges:
        first_center = int(window_size / 2)
        last_center = n_vols - 1 - int(window_size / 2)
    else:
        first_center = 0
        last_center = n_vols - 1
    window_center = first_center
    while window_center <= last_center:
        window_start = window_center - int(window_size / 2)
        if window_start < 0:
            window_start = 0
        window_end = window_center + int(window_size / 2)
        if window_end >= n_vols:
            window_end = n_vols - 1
        window_bold = bold[:, window_start:window_end+1]
        window_center += window_step
        # discard the window if more than `outlier_threshold`
        # (default: 50%) of its volumes are motion outliers
        if (window_bold.sum(axis=0) == 0).mean() > outlier_threshold:
            continue
        window_fc = np.corrcoef(window_bold)
        if exc_interhemispheric:
            # set interhemispheric connections to NaN
            rh_idx = nodes // 2
            window_fc[:rh_idx, rh_idx:] = np.NaN
            window_fc[rh_idx:, :rh_idx] = np.NaN
        window_fcs.append(window_fc[:, :, np.newaxis])
        # calculate lower triangular part of the window FC
        # and drop NaNs (interhemispheric connections)
        # to be used for FCD calculation
        window_fc_tril = window_fcs[-1][np.tril_indices(nodes, -1)]
        window_fc_tril = window_fc_tril[~np.isnan(window_fc_tril), np.newaxis]
        window_fc_trils.append(window_fc_tril)
    # calculate FCD matrix
    window_fc_trils = np.concatenate(window_fc_trils, axis=1)
    fcd_matrix = np.corrcoef(window_fc_trils.T)
    if return_tril:
        fcd_matrix = fcd_matrix[np.tril_indices(fcd_matrix.shape[0], -1)]
    # concatenate window FCs
    window_fcs = np.concatenate(window_fcs, axis=2)
    if return_dfc:
        return fcd_matrix, window_fcs
    else:
        return fcd_matrix

def fcd_ks_device(sim_fcd_trils, emp_fcd_tril, usable_mem=None):
    """
    Calculates Kolmogorov-Smirnov distance of provided simulated FCDs
    to the empirical FCD on GPU. The calculation will be done in batches
    depending on the available GPU memory (`usable_mem`) and the memory
    required for each simulation.

    Parameters
    ----------
    sim_fcd_trils: :obj:`cp.ndarray`
        2D array of simulated FCDs. Shape: (n_simulations, m1)
    emp_fcd_tril: :obj:`cp.ndarray`
        1D array of empirical FCD. Shape: (m2)
    usable_mem: :obj:`int`
        available GPU memory in bytes
        If not provided, 80% of the free GPU memory will be used.
    
    Returns
    -------
    :obj:`np.ndarray`
        Kolmogorov-Smirnov distances. Shape: (n_simulations,)
    """
    if not has_cupy:
        raise RuntimeError(
            "CuPy is not available. Install `cubnm[cupy-cuda11x]` or `cubnm[cupy-cuda12x]`."
        )

    if usable_mem is None:
        free_mem, _ = cp.cuda.runtime.memGetInfo()
        usable_mem = int(free_mem * 0.8)
    n_simulations, m1 = sim_fcd_trils.shape
    m2 = emp_fcd_tril.shape[0]

    # sort emp_fcd_tril and copy to device
    emp_sorted = cp.asarray(emp_fcd_tril, dtype=cp.float64)
    emp_sorted.sort()

    # memory estimation to determine batch size
    mem_emp = emp_sorted.nbytes # invariant to number of simulations
    mem_per_sim = (
        # n_cols * n_bytes * n_variables
        m1 * 8 * 1 + # sim_sorted
        m2 * 8 * 1 + # emp_tiled
        (m1 + m2) * 8 * 3 + # merged_sorted, cdf_sim (also cdf_diff, abs_cdf_diff), cdf_emp
        1 * 8 * 1 # max
    )
    mem_batches = usable_mem - mem_emp
    sims_per_batch = mem_batches // mem_per_sim
    ## this will rarely happen, but just in case:
    assert sims_per_batch > 0, \
        "Not enough GPU memory available for calculating"\
        " FCD KS of even a single simulation."
    n_batches = int(np.ceil(n_simulations / sims_per_batch))
    if n_batches > 1:
        print(f"Calculating FCD KS in {n_batches} batches to reduce GPU memory usage.")
    all_ks = []
    # allocate memory to arrays that are reused across batches
    sim_sorted = cp.zeros((sims_per_batch, m1), dtype=cp.float64)
    merged_sorted = cp.zeros((sims_per_batch, m1 + m2), dtype=cp.float64)
    cdf_sim = cp.zeros((sims_per_batch, m1 + m2), dtype=cp.float64) # also will be used for cdf_diff and abs_cdf_diff
    cdf_emp = cp.zeros((sims_per_batch, m1 + m2), dtype=cp.float64)
    maxes = cp.zeros(sims_per_batch, dtype=cp.float64)
    # repeat emp_sorted for n_batch rows to compare to sim_sorted
    emp_tiled = cp.tile(emp_sorted[None, :], (sims_per_batch, 1))

    for i in range(n_batches):
        # clean up memory before each batch
        gc.collect()
        cp.get_default_memory_pool().free_all_blocks()        
        # calculate batch start and end indices + size
        start = i * sims_per_batch
        end = min((i + 1) * sims_per_batch, n_simulations)
        batch_size = end - start
        # copy sim_fcd_trils to device and sort
        sim_sorted[:batch_size] = cp.asarray(sim_fcd_trils[start:end], cp.float64)
        sim_sorted.sort(axis=1)   
        # merge sim_sorted with emp_sorted and then re-sort
        merged_sorted[:batch_size, :m1] = sim_sorted[:batch_size]
        merged_sorted[:batch_size, m1:] = emp_tiled[:batch_size]
        merged_sorted.sort(axis=1)
        # calculate grid dimensions for kernel launch
        threadsperblock = (16, 16)
        blockspergrid_x = int(np.ceil(batch_size / threadsperblock[0]))
        blockspergrid_y = int(np.ceil((m1 + m2) / threadsperblock[1]))
        blockspergrid = (blockspergrid_x, blockspergrid_y)                
        # launch cdf kernels on GPU to calculate
        # CDFs of sim and emp FCDs for current batch
        cdf_2d_2d_right[blockspergrid, threadsperblock](
            sim_sorted[:batch_size], merged_sorted[:batch_size], cdf_sim[:batch_size]
        )
        cdf_2d_2d_right[blockspergrid, threadsperblock](
            emp_tiled[:batch_size], merged_sorted[:batch_size], cdf_emp[:batch_size]
        )
        cuda.synchronize()
        # calculate KS distance for each simulation as max(abs(cdf_sim - cdf_emp))
        # note: reusing the same array (cdf_sim) for intermediate steps to reduce
        # GPU memory usage
        cp.subtract(cdf_sim[:batch_size], cdf_emp[:batch_size], out=cdf_sim[:batch_size])
        cp.abs(cdf_sim[:batch_size], out=cdf_sim[:batch_size])
        cp.max(cdf_sim[:batch_size], axis=1, out=maxes[:batch_size])
        ks = maxes[:batch_size].get()
        all_ks.append(ks)

    # clean up GPU memory
    del emp_sorted, emp_tiled, sim_sorted, merged_sorted, cdf_emp, cdf_sim, maxes
    gc.collect()
    cp.get_default_memory_pool().free_all_blocks()
    # combine batches and return
    return np.concatenate(all_ks)

def fc_corr_device(sim_fc_trils, emp_fc_tril, usable_mem=None):
    """
    Calculates Pearson correlation between simulated FCs and empirical FCs
    on GPU. The calculation will be done in batches depending on the available
    GPU memory (`usable_mem`) and the memory required for each simulation.

    Parameters
    ----------
    sim_fc_trils: :obj:`cp.ndarray`
        2D array of simulated FCs. Shape: (n_simulations, n_pairs)
    emp_fc_tril: :obj:`cp.ndarray`
        1D array of empirical FC. Shape: (n_pairs)
    usable_mem: :obj:`int`
        available GPU memory in bytes
        If not provided, 80% of the free GPU memory will be used.
    
    Returns
    -------
    :obj:`np.ndarray`
        Pearson correlation coefficients. Shape: (n_simulations,)
    """
    if not has_cupy:
        raise RuntimeError(
            "CuPy is not available. Install `cubnm[cupy-cuda11x]` or `cubnm[cupy-cuda12x]`."
        )

    if usable_mem is None:
        free_mem, _ = cp.cuda.runtime.memGetInfo()
        usable_mem = int(free_mem * 0.8)
    n_simulations, n_pairs = sim_fc_trils.shape

    # copy empirical FC to device and z-score
    # note that ddof=1 is necessary for the same
    # results as scipy.stats.pearsonr
    emp_fc = cp.asarray(emp_fc_tril[None, :])
    mean = emp_fc.mean(axis=1, keepdims=True)
    sd = emp_fc.std(axis=1, ddof=1, keepdims=True)
    # z-scoring is done in-place to reduce memory usage
    emp_fc -= mean
    emp_fc /= sd

    # memory estimation to determine batch size
    mem_emp = emp_fc.nbytes
    mem_per_sim = (
        # n_cols * n_bytes * n_variables
        n_pairs * 8 * 1 + # sim_fc
        1 * 8 * 3 # mean and sd and corr
    )
    mem_batches = usable_mem - mem_emp
    sims_per_batch = max(1, mem_batches // mem_per_sim)
    ## this will rarely happen, but just in case:
    assert sims_per_batch > 0, \
        "Not enough GPU memory available for calculating"\
        " FCD KS of even a single simulation."
    n_batches = int(np.ceil(n_simulations / sims_per_batch))
    if n_batches > 1:
        print(f"Calculating FC corr in {n_batches} batches to reduce GPU memory usage.")

    all_corr = []
    # allocate memory to arrays that are reused across batches
    sim_fc = cp.zeros((sims_per_batch, n_pairs), dtype=cp.float64)
    corr = cp.zeros((sims_per_batch, 1), dtype=cp.float64)
    mean = cp.zeros((sims_per_batch, 1), dtype=cp.float64)
    sd = cp.zeros((sims_per_batch, 1), dtype=cp.float64)
    
    for i in range(n_batches):
        # clean up memory before each batch
        gc.collect()
        cp.get_default_memory_pool().free_all_blocks()        
        # calculate batch start and end indices + size
        start = i * sims_per_batch
        end = min((i + 1) * sims_per_batch, n_simulations)
        batch_size = end - start
        # zscore sim_fc
        sim_fc[:batch_size] = cp.asarray(sim_fc_trils[start:end], cp.float64)
        sim_fc[:batch_size].mean(axis=1, keepdims=True, out=mean[:batch_size])
        sim_fc[:batch_size].std(axis=1, ddof=1, keepdims=True, out=sd[:batch_size])
        sim_fc[:batch_size] -= mean[:batch_size]
        sim_fc[:batch_size] /= sd[:batch_size]
        # matrix multiplication to calculate correlation
        # divide by n_pairs-1 to get unbiased estimate
        # (following scipy.stats.pearsonr)
        cp.matmul(sim_fc[:batch_size], emp_fc.T, out=corr[:batch_size])
        corr[:batch_size] /= n_pairs-1
        all_corr.append(corr[:batch_size].get().flatten())

    # clean up GPU memory
    del sim_fc, emp_fc, corr, mean, sd
    gc.collect()
    cp.get_default_memory_pool().free_all_blocks()
    # combine batches and return
    return np.concatenate(all_corr)


# Autogenerated using:
# - Template: 'src/cubnm/sim/simgroup.py.mako'
# - Model definition: 'codegen/recipes/rww.yaml'
# Do not modify this autogenerated code. Instead modify template
# or model definition.
from .base import SimGroup
import sys
import logging
import numpy as np

logging.basicConfig(stream=sys.stdout, level=logging.WARNING)
logger = logging.getLogger(__name__)

class rWWSimGroup(SimGroup):
    model_name = "rWW"
    global_param_names = ['G']
    regional_param_names = ['w_p', 'J_N', 'wIE', 'sigma']
    state_names = ['I_E', 'I_I', 'r_E', 'r_I', 'S_E', 'S_I']
    sel_state_var = "r_E"
    n_noise = 2
    default_params = {
        "G": None,  # required
        "w_p": 1.4,
        "J_N": 0.15,
        "wIE": 1.0,
        "sigma": 0.01,
    }

    def __init__(
        self, 
        *args, 
        do_fic = True, 
        max_fic_trials = 0, 
        I_SAMPLING_START = 1000, 
        I_SAMPLING_END = 10000, 
        init_delta = 0.02, 
        fic_penalty_scale = 0.5, 
        **kwargs
    ):
        r"""
        Group of reduced Wong-Wang (excitatory-inhibitory) simulations that are executed in parallel.

        Parameters
        ---------
        do_fic: :obj:`bool`
            whether to apply feedback inhibition control. If provided wIE parameters will be ignored
        max_fic_trials: :obj:`int`
            maximum number of numerical FIC trials. If set to 0, FIC will be done only analytically
        I_SAMPLING_START: :obj:`int`
            starting time of numerical FIC I_E sampling (msec)
        I_SAMPLING_END: :obj:`int`
            end time of numerical FIC I_E sampling (msec)
        init_delta: :obj:`float`
            initial delta for numerical FIC adjustment
        fic_penalty_scale: :obj:`float`
            how much deviation from FIC target mean rE of 3 Hz is penalized. Set to 0 to disable FIC penalty.
        *args, **kwargs:
            see :class:`cubnm.sim.SimGroup` for details

        Attributes
        ----------
        param_lists: :obj:`dict` of :obj:`np.ndarray`
            dictionary of parameter lists, including
                - ``'G'``: global coupling strength. Shape: (N_SIMS,)
                - ``'w_p'``: local excitatory recurrence. Shape: (N_SIMS, nodes)
                - ``'J_N'``: synaptic coupling strength. Shape: (N_SIMS, nodes)
                - ``'wIE'``: inhibitory to excitatory weight. Shape: (N_SIMS, nodes)
                - ``'sigma'``: noise amplitude. Shape: (N_SIMS, nodes)
                - ``'v'``: conduction velocity. Shape: (N_SIMS,)

        Equations
        ---------
        .. math::

            \begin{gather}
            I_i^E=W^EI_b+w_i^{p} J_i^{N} S_i^E+GJ_i^{N}\sum_{j}{C_{ij}S_j^E}-w_i^{IE}S_i^I \\
            I_i^I=W^II_b+J_i^{N}S_i^E-w^{II}S_i^I \\
            r_i^E=H^E(I_i^E)\ =\frac{a^EI_i^E-b^E}{1\ -\ e^{-d^E(a^EI_i^E-b^E)}} \\
            r_i^I=H^I(I_i^I)\ =\frac{a^II_i^I-b^I}{1\ -\ e^{-d^I(a^II_i^I-b^I)}} \\
            \dot{S_i^E}=-\frac{S_i^E}{\tau_E}+(1\ -\ S_i^E)\gamma r_i^E(t)+\sigma_i\epsilon_i^E \\
            \dot{S_i^I}=-\frac{S_i^I}{\tau_I}+r_i^I+\sigma_i\epsilon_i^I \\
            \end{gather}

        References
        ----------        
        * Deco et al. 2014 Journal of Neuroscience (10.1523/JNEUROSCI.5068-13.2014)
        * Demirtas et al. 2019 Neuron (10.1016/j.neuron.2019.01.017)
        * Deco et al. 2018 Current Biology (10.1016/j.cub.2018.07.083)
        """
        self.do_fic = do_fic
        self.max_fic_trials = max_fic_trials
        self.I_SAMPLING_START = I_SAMPLING_START
        self.I_SAMPLING_END = I_SAMPLING_END
        self.init_delta = init_delta
        self.fic_penalty_scale = fic_penalty_scale
        # parent init must be called after setting
        # model-specific attributes because
        # it sets .last_config which may include some
        # model-specific attributes
        super().__init__(*args, **kwargs)

    def get_config(self, *args, **kwargs):
        config = super().get_config(*args, **kwargs)
        config.update({
            'do_fic': self.do_fic,
            'max_fic_trials': self.max_fic_trials,
            'I_SAMPLING_START': self.I_SAMPLING_START,
            'I_SAMPLING_END': self.I_SAMPLING_END,
            'init_delta': self.init_delta,
            'fic_penalty_scale': self.fic_penalty_scale,
        })
        return config
    
    @property
    def _model_config(self):
        """
        Internal model configuration used in the simulation
        """
        model_config = super()._model_config
        model_config.update({
            'do_fic': str(int(self.do_fic)),
            'max_fic_trials': str(self.max_fic_trials),
            'I_SAMPLING_START': str(self.I_SAMPLING_START),
            'I_SAMPLING_END': str(self.I_SAMPLING_END),
            'init_delta': str(self.init_delta),
            'fic_penalty_scale': str(self.fic_penalty_scale),
        })
        return model_config

    @property
    def labels(self):
        """
        Labels of parameters and state variables
        to use in plots and reports
        """
        labels = super().labels
        labels.update({
            'G': r'G',
            'w_p': r'$w^{p}$',
            'J_N': r'$J^{N}$',
            'wIE': r'$w^{IE}$',
            'sigma': r'$\sigma$',
            'I_E': r'$I^E$',
            'I_I': r'$I^I$',
            'r_E': r'$r^E$',
            'r_I': r'$r^I$',
            'S_E': r'$S^E$',
            'S_I': r'$S^I$',
        })
        return labels
    
    @property
    def _to_clear(self):
        """
        List of attributes to be cleared when clear() is called
        """
        attrs = super()._to_clear + [
            "fic_unstable",
            "fic_failed",
            "fic_ntrials",
        ]
        return [attr for attr in attrs if hasattr(self, attr)]

    @SimGroup.N.setter
    def N(self, N):
        super(rWWSimGroup, rWWSimGroup).N.__set__(self, N)
        if self.do_fic:
            self.param_lists["wIE"] = np.zeros((self._N, self.nodes), dtype=float)

    def _process_out(self, out):
        super()._process_out(out)
        if self.do_fic:
            # FIC stats
            self.fic_unstable = self._global_bools[0]
            self.fic_failed = self._global_bools[1]
            self.fic_ntrials = self._global_ints[0]
            # write FIC (+ numerical adjusted) wIE to param_lists
            self.param_lists["wIE"] = self._regional_params[2].reshape(self.N, -1)

    def score(self, emp_fc_tril=None, emp_fcd_tril=None, emp_bold=None, **kwargs):
        """
        Calculates individual goodness-of-fit terms and aggregates them.
        In FIC models also calculates fic_penalty. Note that while 
        FIC penalty is included in the cost function of optimizer, 
        it is not included in the aggregate GOF.

        Parameters
        --------
        emp_fc_tril: :obj:`np.ndarray`
            1D array of empirical FC lower triangle. Shape: (edges,)
        emp_fcd_tril: :obj:`np.ndarray`
            1D array of empirical FCD lower triangle. Shape: (window_pairs,)
        emp_bold: :obj:`np.ndarray` or None
            cleaned and parcellated empirical BOLD time series. Shape: (nodes, volumes)
            Motion outliers should either be excluded (not recommended as it disrupts
            the temporal structure) or replaced with zeros.
            If provided emp_fc_tril and emp_fcd_tril will be ignored.
        **kwargs:
            keyword arguments passed to `cubnm.sim.SimGroup.score`

        Returns
        -------
        scores: :obj:`pd.DataFrame`
            The goodness of fit measures (columns) of each simulation (rows)
        """
        scores = super().score(emp_fc_tril, emp_fcd_tril, emp_bold, **kwargs)
        # calculate FIC penalty
        if self.do_fic & (self.fic_penalty_scale > 0):
            # calculate time-averaged r_E in each simulation-node
            if self.states_ts:
                mean_r_E = self.sim_states["r_E"][:, self.n_states_samples_remove:].mean(axis=1)
            else:
                mean_r_E = self.sim_states["r_E"]
            for idx in range(self.N):
                # absolute deviation from target of 3 Hz
                diff_r_E = np.abs(mean_r_E[idx, :] - 3)
                if (diff_r_E > 1).sum() > 0:
                    # at least one node has a deviation greater than 1
                    # only consider deviations greater than 1
                    # in the penalty
                    diff_r_E[diff_r_E <= 1] = np.NaN
                    # within each node the FIC penalty decreases
                    # exponentially as the deviations gets smaller
                    # the penalty max value in each node is 1
                    # which is then scaled by fic_penalty_scale
                    # and averaged across nodes (with 0s/NaNs)
                    # in node with no deviation > 1 Hz
                    scores.loc[idx, "-fic_penalty"] = (
                        -np.nansum(1 - np.exp(-0.05 * (diff_r_E - 1)))
                        * self.fic_penalty_scale / self.nodes
                    )
                else:
                    scores.loc[idx, "-fic_penalty"] = 0
        return scores

    def _get_save_data(self):
        """
        Get the simulation outputs and parameters to be saved to disk

        Returns
        -------
        out_data: :obj:`dict`
            dictionary of simulation outputs and parameters
        """
        out_data = super()._get_save_data()
        if self.do_fic:
            out_data.update(
                fic_unstable=self.fic_unstable,
                fic_failed=self.fic_failed,
                fic_ntrials=self.fic_ntrials,
            )
        return out_data

    def _problem_init(self, problem):
        """
        Extends BNMProblem initialization and includes FIC penalty
        if indicated.

        Parameters
        ----------
        problem: :obj:`cubnm.optimize.BNMProblem`
            optimization problem object
        """
        if (self.do_fic) & ("wIE" in problem.het_params):
            raise ValueError(
                "In rWW wIE should not be specified as a heterogeneous parameter when FIC is done"
            )
        if problem.multiobj:
            if self.do_fic & (self.fic_penalty_scale > 0):
                problem.obj_names.append("-fic_penalty")
                problem.n_obj += 1

    def _problem_evaluate(self, problem, X, out, *args, **kwargs):
        """
        Extends BNMProblem evaluation and includes FIC penalty
        in the cost function if indicated.

        Parameters
        ----------
        X: :obj:`np.ndarray`
            the normalized parameters of current population in range [0, 1]. 
            Shape: (N, ndim)
        out: :obj:`dict`
            the output dictionary to store the results with keys ``'F'`` and ``'G'``.
            Currently only ``'F'`` (cost) is used.
        *args, **kwargs
        """
        # Note: scores (inidividual GOF measures) is passed on in kwargs 
        # in the internal mechanism of pymoo evaluation function
        scores = kwargs["scores"][-1]
        if self.do_fic & (self.fic_penalty_scale > 0):
            if problem.multiobj:
                out["F"] = np.concatenate(
                    [out["F"], -scores.loc[:, ["-fic_penalty"]].values], axis=1
                )
            else:
                out["F"] -= scores.loc[:, "-fic_penalty"].values

    @classmethod
    def _get_test_configs(cls, cpu_gpu_identity=False):
        """
        Get configs for testing the simulations

        Parameters
        ----------
        cpu_gpu_identity: :obj:`bool`
            indicates whether configs are for CPU/GPU identity tests
            in which case force_cpu is not included in the configs
            since tests will be done on both CPU and GPU

        Returns
        -------
        configs: :obj:`dict` of :obj:`list`
        """
        configs = super()._get_test_configs(cpu_gpu_identity)
        configs.update({
            # 0: no FIC
            # 1: analytical FIC
            # 2: analytical + numerical FIC
            'do_fic': [0, 1, 2]
        })
        return configs

    @classmethod
    def _get_test_instance(cls, opts):
        """
        Initializes an instance that is used in tests

        Parameters
        ----------
        opts: :obj:`dict`
            dictionary of test options

        Returns
        -------
        sim_group: :obj:`cubnm.sim.rWWSimGroup`
            simulation group object of the test simulation
            which is not run yet
        """
        # initialze sim group
        sim_group = super()._get_test_instance(opts)
        # set do_fic
        sim_group.do_fic = opts['do_fic'] > 0
        if opts['do_fic'] == 2:
            sim_group.max_fic_trials = 5
        return sim_group



"""
Example datasets
"""
import sys
import os
import logging
import numpy as np
import pandas as pd

if sys.version_info >= (3, 9):
    from importlib.resources import files
else:
    from importlib_resources import files

from cubnm import utils

logging.basicConfig(stream=sys.stdout, level=logging.WARNING)
logger = logging.getLogger(__name__)

def load_sc(
        measure, 
        parc="schaefer-100", 
        sub="group-train706", 
        norm="mean",
    ):
    """
    Load example structural connectivity matrix from HCP-YA dataset

    Parameters
    ----------
    measure: {'strength', 'length'}
        Structural connectivity measure.

        - ``'strength'``: SC strength (tract counts)
        - ``'length'``: SC tracts length

    parc: :obj:`str`
        Parcellation. Currently only ``'schaefer-100'`` is supported.
    sub: :obj:`str`
        Subject or group ID. Two subjects and two group-averaged 
        matrices are currently included:

        - ``'100206'``
        - ``'100307'``
        - ``'group-train706'``
        - ``'group-test303'``

    norm: {'mean', None}
        SC strength normalization method.
        Only used when ``measure`` is ``'strength'``.

        - ``'mean'``: normalize to mean 0.01. Default
        - ``None``: no normalization

    Returns
    -------
    :obj:`np.ndarray`
        Structural connectivity matrix. Shape: (nodes, nodes)
    """
    path = (
        files("cubnm.data").joinpath('hcp')
        .joinpath('sc').joinpath(sub)
        .joinpath(f'ctx_parc-{parc}_desc-{measure}.npz')
    ).as_posix()
    try:
        mat = np.load(path)['arr_0']
    except FileNotFoundError as e:
        raise ValueError(
            f"Structural connectivity {measure} matrix in {parc} "
            f"for {sub} is not included in the toolbox."
        ) from e
    # normalization
    if measure == 'strength':
        if norm == 'mean':
            mat /= (mat.mean() * 100)
    return mat

def load_bold(parc="schaefer-100", sub="100206", ses="REST1_LR"):
    """
    Load example BOLD data from the HCP-YA dataset

    Parameters
    --------
    parc: :obj:`str`
        Parcellation. Currently only ``'schaefer-100'`` is available.
    sub: :obj:`str`
        Subject ID. Currently two subjects are included:

        - ``'100206'``
        - ``'100307'``

    ses: {'REST1_LR', 'REST2_LR'}
        Imaging session.

    Returns
    --------
    :obj:`np.ndarray`
        The BOLD signal. Shape: (nodes, volumes)
    """
    path = (
        files("cubnm.data").joinpath('hcp').joinpath('bold')
        .joinpath(sub).joinpath(ses)
        .joinpath(f'ctx_parc-{parc}_desc-bold.npz')
    ).as_posix()
    try:
        out = np.load(path)['arr_0']
    except FileNotFoundError as e:
        raise ValueError(
            f"BOLD data for {parc} and {sub} in session {ses} "
            f"is not included in the toolbox."
        ) from e
    return out

def load_fc(
        parc="schaefer-100", 
        sub="group-train706", 
        ses="REST",
        exc_interhemispheric=False,
        return_tril=True,
    ):
    """
    Load example FC from the HCP-YA dataset

    Parameters
    --------
    parc: :obj:`str`
        Parcellation. Currently only ``'schaefer-100'`` is available.
    sub: :obj:`str`
        Subject or group ID. Two subjects and two group-averaged 
        matrices are currently included:

        - ``'100206'``
        - ``'100307'``
        - ``'group-train706'``
        - ``'group-test303'``

    ses: {'REST1_LR', 'REST2_LR', 'REST'}
        Imaging session. For subject-level data currently
        only ``'REST1_LR'``, and ``'REST2_LR'`` are included and
        for group-level data only ``'REST'`` is included.
        ``'REST'`` includes the data averaged across
        all sessions.
    exc_interhemispheric: :obj:`bool`
        Whether to exclude interhemispheric connections. Default: False
    return_tril: :obj:`bool`
        Whether to return the lower triangular part of the FC matrix.

    Returns
    --------
    :obj:`np.ndarray`
        Functional connectivity matrix or its lower triangular part.
        Shape: (nodes, nodes) or (node_pairs,)
    """
    if "group-" in sub:
        if exc_interhemispheric:
            tags = '_exc-inter'
        else:
            tags = ''
        path = (
            files("cubnm.data").joinpath('hcp').joinpath('fc')
            .joinpath(sub).joinpath(ses)
            .joinpath(f'ctx_parc-{parc}{tags}_desc-fc.npz')
        ).as_posix()
        try:
            fc = np.load(path)['arr_0']
        except FileNotFoundError as e:
            raise ValueError(
                f"Functional connectivity matrix for {parc} "
                f"and {sub} in session {ses} is not included in the toolbox."
            ) from e
        if return_tril:
            fc = fc[np.tril_indices(fc.shape[0], -1)]
            if exc_interhemispheric:
                fc = fc[~np.isnan(fc)]
        return fc
    else:
        # load bold
        bold = load_bold(
            parc=parc, 
            sub=sub,
            ses=ses, 
        )
        # calculate FC
        fc = utils.calculate_fc(
            bold, 
            exc_interhemispheric=exc_interhemispheric,
            return_tril=return_tril
        )
        return fc

def load_fcd(
        parc="schaefer-100", 
        sub="group-train706",
        ses="REST", 
        window_size=30,
        window_step=5,
        drop_edges=True,
        exc_interhemispheric=False,
        return_tril=True,
    ):
    """
    Load example FCD from the HCP-YA dataset

    Parameters
    --------
    parc: :obj:`str`
        Parcellation. Currently only ``'schaefer-100'`` is available.
    sub: :obj:`str`
        Subject or group ID. Two subjects and two group-averaged 
        matrices are currently included:

        - ``'100206'``
        - ``'100307'``
        - ``'group-train706'``
        - ``'group-test303'``

    ses: {'REST1_LR', 'REST2_LR', 'REST'}
        Imaging session. For subject-level data currently
        only ``'REST1_LR'``, and ``'REST2_LR'`` are included and
        for group-level data only ``'REST'`` is included.
        ``'REST'`` includes the data pooled across
        all sessions.
    window_size: :obj:`int`
        dynamic FC window size (in seconds)
        will be converted to N TRs (nearest even number)
        The actual window size is number of TRs + 1 (including center)
    window_step: :obj:`int`
        dynamic FC window step (in seconds)
        will be converted to N TRs
    drop_edges: :obj:`bool`
        drop edge windows which have less than window_size volumes
    exc_interhemispheric: :obj:`bool`
        Whether to exclude interhemispheric connections. Default: False
    return_tril: :obj:`bool`
        Whether to return the lower triangular part of the FCD matrix.
        Ignored when ``sub`` is a group (in this case
        the return matrix is pooled from FCD lower triangles of
        individual subjects).

    Returns
    --------
    :obj:`np.ndarray`
        Functional connectivity dynamics matrix or its lower triangular part.
        Shape: (nodes, nodes) or (node_pairs,)
    """
    if "group-" in sub:
        if not return_tril:
            logger.warning(
                "return_tril is ignored when sub='group-all'. "
                "Returning lower triangle."
            )
        if exc_interhemispheric:
            tags = '_exc-inter'
        else:
            tags = ''
        tags += f'_window-{window_size}_step-{window_step}'
        path = (
            files("cubnm.data").joinpath('hcp').joinpath('fcd')
            .joinpath(sub).joinpath(ses)
            .joinpath(f'ctx_parc-{parc}{tags}_desc-fcdtril.npz')
        ).as_posix()
        try:
            fcd = np.load(path)['arr_0']
        except FileNotFoundError as e:
            raise ValueError(
                f"Functional connectivity dynamics for {parc} "
                f"and {sub} in session {ses} using window size {window_size} "
                f"and step {window_step} is not included in the toolbox."
            ) from e
        return fcd
    else:
        # create it from bold
        bold = load_bold(
            parc=parc, 
            sub=sub, 
            ses=ses
        )
        # convert window size and step to TRs
        TR = 0.72
        window_size_TRs = int(np.round(window_size / (TR*2))) * 2
        window_step_TRs = int(np.round(window_step / TR))
        # calculate FCD
        fcd = utils.calculate_fcd(
            bold, 
            window_size=window_size_TRs, 
            window_step=window_step_TRs,
            drop_edges=drop_edges,
            exc_interhemispheric=exc_interhemispheric,
            return_tril=return_tril
        )
        return fcd

def load_maps(names, parc="schaefer-100", norm="minmax"):
    """
    Loads example heterogeneity maps

    Parameters
    ----------
    names: :obj:`str` or :obj:`list`
        One or more maps selected from this list:

        - ``'myelinmap'``
        - ``'thickness'``
        - ``'fcgradient01'``
        - ``'genepc1'``
        - ``'nmda'``
        - ``'gabaa'``
        - ``'yeo7'``

    parc: :obj:`str`
        Parcellation. Currently only ``'schaefer-100'`` is supported.
    norm: {'zscore', 'minmax', None}
        Map normalization method applied across nodes.

        - ``'zscore'``: maps are z-score normalized
        - ``'minmax'``: maps are min-max normalized to [0, 1]
        - ``None``: no normalization

    Returns
    --------
    :obj:`np.ndarray` or :obj:`str`
        Maps arrays or path to their
        text file. Shape: (maps, nodes)

    Notes
    -----
    For more information and code on how these maps were
    obtained and parcellated see ``utils.datasets.load_maps``
    in https://github.com/amnsbr/eidev. The set of maps included
    here are limited and provided just as examples. We recommend 
    users to use ``neuromaps`` and similar tools to obtain and 
    parcellate further maps.
    """
    # TODO: Add other parcellations
    if isinstance(names, str):
        names = [names]
    maps = []
    for name in names:
        if name not in ['yeo7']:
            filename = f"ctx_parc-{parc}_desc-{name}_zscore.txt"
        else:
            norm = None
            filename = f"ctx_parc-{parc}_desc-{name}.txt"
        path = files("cubnm.data").joinpath('maps').joinpath(filename).as_posix()
        curr_map = np.loadtxt(path)
        if norm=='minmax':
            curr_map = (curr_map - np.min(curr_map)) / (np.max(curr_map) - np.min(curr_map))
        maps.append(curr_map.reshape(1, -1))
    maps = np.vstack(maps)
    return maps



# Autogenerated using:
# - Template: 'src/cubnm/sim/simgroup.py.mako'
# - Model definition: 'codegen/recipes/kuramoto.yaml'
# Do not modify this autogenerated code. Instead modify template
# or model definition.
from .base import SimGroup
import sys
import logging
import numpy as np

logging.basicConfig(stream=sys.stdout, level=logging.WARNING)
logger = logging.getLogger(__name__)

class KuramotoSimGroup(SimGroup):
    model_name = "Kuramoto"
    global_param_names = ['G']
    regional_param_names = ['init_theta', 'omega', 'sigma']
    state_names = ['theta']
    sel_state_var = "theta"
    n_noise = 1
    default_params = {
        "G": None,  # required
        "init_theta": None,  # required
        "omega": np.pi,
        "sigma": 0.17,
    }

    def __init__(
        self, 
        *args, 
        random_init_theta = True, 
        **kwargs
    ):
        r"""
        Group of Kuramoto simulations that are executed in parallel.

        Parameters
        ---------
        random_init_theta: :obj:`bool`
            set initial theta by randomly sampling from a uniform distribution
        *args, **kwargs:
            see :class:`cubnm.sim.SimGroup` for details

        Attributes
        ----------
        param_lists: :obj:`dict` of :obj:`np.ndarray`
            dictionary of parameter lists, including
                - ``'G'``: global coupling strength. Shape: (N_SIMS,)
                - ``'init_theta'``: initial phase angle. Shape: (N_SIMS, nodes)
                - ``'omega'``: natural frequency. Shape: (N_SIMS, nodes)
                - ``'sigma'``: noise amplitude. Shape: (N_SIMS, nodes)
                - ``'v'``: conduction velocity. Shape: (N_SIMS,)

        Equations
        ---------
        .. math::

            \begin{gather}
            \dot{\theta_i} = \omega_i + G \sum_j{C_{ij} \sin(\theta_j - \theta_i)} + \sigma_i\epsilon_i \\
            \end{gather}

        References
        ----------        
        * Kuramoto et al. 1984 "Chemical Oscillations, Waves, and Turbulence"
        * Cabral et al. 2011 NeuroImage (10.1016/j.neuroimage.2011.04.010)

        Notes
        -----
        In Kuramoto model GPU and CPU results may differ under
        certain conditions (large G, presence of conduction delay, large
        sigma, etc.), especially when using larger dt.

        This discrepancy arises from subtle variations in mathematical
        function implementations between GPU and CPU hardware, which,
        depending on the system dynamics, can amplify through butterfly effects.

        As noted in NVIDIA's documentation
        (https://docs.nvidia.com/cuda/floating-point/index.html#considerations-for-a-heterogeneous-world),
        such differences do not indicate that either computation is incorrect.

        We therefore recommend to assess sensitivity of simulated
        findings in Kuramoto model (e.g. optimal simulation) to 
        CPU-based versus GPU-based simulations.
        """
        self.random_init_theta = random_init_theta
        # parent init must be called after setting
        # model-specific attributes because
        # it sets .last_config which may include some
        # model-specific attributes
        super().__init__(*args, **kwargs)

    def get_config(self, *args, **kwargs):
        config = super().get_config(*args, **kwargs)
        config.update({
            'random_init_theta': self.random_init_theta,
        })
        return config
    
    @property
    def _model_config(self):
        """
        Internal model configuration used in the simulation
        """
        model_config = super()._model_config
        model_config.update({
            'random_init_theta': str(int(self.random_init_theta)),
        })
        return model_config

    def post_init(self):
        logger.warning(
            "In Kuramoto model GPU and CPU results may differ under "
            "certain conditions (large G, presence of conduction delay, large "
            "sigma, etc.), especially when using larger dt.\n"
            "This discrepancy arises from subtle variations in mathematical "
            "function implementations between GPU and CPU hardware, which, "
            "depending on the system dynamics, can amplify through butterfly effects.\n"
            "As noted in NVIDIA's documentation "
            "(https://docs.nvidia.com/cuda/floating-point/index.html#considerations-for-a-heterogeneous-world), "
            "such differences do not indicate that either computation is incorrect.\n"
            "We therefore recommend to assess sensitivity of simulated "
            "findings in Kuramoto model (e.g. optimal simulation) to "
            "CPU-based versus GPU-based simulations."
        )

    @SimGroup.N.setter
    def N(self, N):
        super(KuramotoSimGroup, KuramotoSimGroup).N.__set__(self, N)
        if self.random_init_theta:
            # sample from uniform distribution of [0, 2*pi] across nodes and
            # repeat it across simulations
            # use the same random seed as the simulation noise
            rng = np.random.default_rng(self.sim_seed)
            self.param_lists["init_theta"] = np.tile(rng.uniform(0, 2 * np.pi, self.nodes), (self._N, 1))
        else:
            self.param_lists["init_theta"] = np.zeros((self._N, self.nodes), dtype=float)
            logger.warning("init_theta is set to zero")



"""
Simulation of the models
"""
import os
import gc
import copy
from decimal import Decimal
import multiprocessing
import sys
import logging
import numpy as np
import scipy.stats
import pandas as pd

from cubnm._core import run_simulations, set_const
from cubnm._setup_opts import (
    many_nodes_flag, gpu_enabled_flag, 
    max_nodes_reg, max_nodes_many,
    noise_segment_flag
)
from cubnm import utils, datasets
from .. import _version
__version__ = _version.get_versions()['version']

try:
    import cupy as cp
    has_cupy = True
except ImportError:
    cp = None
    has_cupy = False

logging.basicConfig(stream=sys.stdout, level=logging.WARNING)
logger = logging.getLogger(__name__)

# NaN values in the scores are replaced with the worst possible scores
# listed below.
WORST_SCORES = {
    "+fc_corr": -1.0,
    "-fc_diff": -2.0,
    "-fc_normec": -1.0,
    "-fcd_ks": -1.0,
}

class SimGroup:
    def __init__(
        self,
        duration,
        TR,
        sc,
        sc_dist=None,
        out_dir=None,
        dt='0.1',
        bw_dt='1.0',
        states_ts=False,
        states_sampling=None,
        noise_out=False,
        do_fc=True,
        do_fcd=True,
        window_size=30,
        window_step=5,
        sim_seed=0,
        exc_interhemispheric=False,
        force_cpu=False,
        force_gpu=False,
        gof_terms=["+fc_corr", "-fcd_ks"],
        bw_params="friston2003",
        bold_remove_s=30,
        fcd_drop_edges=True,
        noise_segment_length=30,
        sim_verbose=False,
        progress_interval=500
    ):
        """
        Group of simulations that are executed in parallel
        (as possible depending on the available hardware)
        on GPU/CPU.

        Parameters
        ---------
        duration: :obj:`float`
            simulation duration (in seconds)
            It must not have more than 3 decimal places.
        TR: :obj:`float`
            BOLD TR (in seconds)
        sc: :obj:`str` or :obj:`np.ndarray`
            path to structural connectome strengths (as an unlabled .txt/.npy)
            or a numpy array. Shape: (nodes, nodes)
            If asymmetric, rows are sources and columns are targets.
        sc_dist: :obj:`str` or :obj:`np.ndarray`
            path to structural connectome distances (as an unlabled .txt)
            or a numpy array. Shape: (nodes, nodes)
            If provided ``'v'`` (velocity) will be a free parameter and there
            will be delay in inter-regional connections.
            If asymmetric, rows are sources and columns are targets.
        out_dir: {:obj:`str` or 'same' or None}
            output directory

            - :obj:`str`: will create a directory in the provided path
            - ``'same'``: will create a directory named based on sc
                (only should be used when sc is a path and not a numpy array)
            - ``None``: will not have an output directory (and cannot save outputs)

        dt: :obj:`decimal.Decimal` or :obj:`str`
            model integration time step (in msec)
        bw_dt: :obj:`decimal.Decimal` or :obj:`str`
            Ballon-Windkessel integration time step (in msec)
        states_ts: :obj:`bool`
            return time series of model states to self.sim_states
            Note that this will increase the memory usage and is not
            recommended for large number of simulations (e.g. in a grid search)
        states_sampling: :obj:`float`
            sampling rate of model states in seconds.
            Default is None, which uses BOLD TR as the sampling rate.
        noise_out: :obj:`bool`
            return noise time series
        do_fc: :obj:`bool`
            calculate simulated functional connectivity (FC)
        do_fcd: :obj:`bool`
            calculate simulated functional connectivity dynamics (FCD)
        window_size: :obj:`int`
            dynamic FC window size (in seconds)
            will be converted to N TRs (nearest even number)
            The actual window size is number of TRs + 1 (including center)
        window_step: :obj:`int`
            dynamic FC window step (in seconds)
            will be converted to N TRs
        sim_seed: :obj:`int`
            seed used for the noise simulation
        exc_interhemispheric: :obj:`bool`
            excluded interhemispheric connections from sim FC and FCD calculations
        force_cpu: :obj:`bool`
            use CPU for the simulations (even if GPU is available). If set
            to False the program might use GPU or CPU depending on GPU
            availability
        force_gpu: :obj:`bool`
            on some HPC/HTC systems occasionally GPUtil might not detect an 
            available GPU device. Use this if there is a GPU available but
            is not being used for the simulation. If set to True but a GPU
            is not available will lead to errors.
        gof_terms: :obj:`list` of :obj:`str`
            list of goodness-of-fit terms to be used for scoring. May include:

            - ``'-fcd_ks'``: negative Kolmogorov-Smirnov distance of FCDs
            - ``'+fc_corr'``: Pearson correlation of FCs
            - ``'-fc_diff'``: negative absolute difference of FC means
            - ``'-fc_normec'``: negative Euclidean distance of FCs \
                divided by max EC [sqrt(n_pairs*4)]

        bw_params: {'friston2003' or 'heinzle2016-3T' or :obj:`dict`}
            see :func:`cubnm.utils.get_bw_params` for details
        bold_remove_s: :obj:`float`
            remove the first bold_remove_s seconds from the simulated BOLD
            in FC, FCD and mean state calculations (but the entire BOLD will
            be returned to .sim_bold)
        fcd_drop_edges: :obj:`bool`
            drop the edge windows in FCD calculations
        noise_segment_length: :obj:`float` or None
            in seconds, length of the noise segments in the simulations
            The noise segment will be repeated after shuffling
            of nodes and time points. To generate noise for the entire
            simulation without repetition, set this to None.
            Note that varying the noise segment length will result
            in a different noise array even if seed is fixed (but
            fixed combination of seed and noise_segment_length will
            result in reproducible noise)
        sim_verbose: :obj:`bool`
            verbose output of the simulation including details of
            simulations and a progress bar.
        progress_interval: :obj:`int`
            msec; interval of progress updates in the simulation
            Only used if ``sim_verbose`` is ``True``

        Attributes
        ---------
        param_lists: :obj:`dict` of :obj:`np.ndarray`
            dictionary of parameter lists, including

            - global parameters with shape (N_SIMS,)
            - regional parameters with shape (N_SIMS, nodes)
            - ``'v'``: conduction velocity. Shape: (N_SIMS,)

        Additional attributes will be added after running the simulations.
        See :func:`cubnm.sim.base.SimGroup._process_out` for details.

        Notes
        -----
        Derived classes must set the following attributes:
            model_name: :obj:`str`
                name of the model used in the simulations
            global_param_names: :obj:`list` of :obj:`str`
                names of global parameters
            regional_param_names: :obj:`list` of :obj:`str`
                names of regional parameters
            state_names: :obj:`list` of :obj:`str`
                names of the state variables
            sel_state_var: :obj:`str`
                name of the state variable used in the tests
            n_noise: :obj:`int`
                number of noise elements per node per time point
                (e.g. 2 if there are noise to E and I neuronal populations)
            default_params: :obj:`dict`
                default parameter values for the simulations.
                If a parameter's default is set to None,
                it will be a required (fixed or free) parameter.
                This is often the case with 'G' (global coupling).
        """
        self.input_sc = sc
        if isinstance(sc, (str, os.PathLike)):
            if sc.endswith(".txt"):
                self.sc = np.loadtxt(self.input_sc)
            elif sc.endswith(".npy"):
                self.sc = np.load(self.input_sc)
        else:
            self.sc = self.input_sc
        if np.any(np.diag(self.sc)):
            logger.warning(
                "The diagonal of the SC matrix is not 0. "
                "Self-connections are not ignored by default in "
                "the simulations. If you want to ignore them set "
                "the diagonal of the SC matrix to 0."
            )
        self.duration = duration
        self.TR = TR
        self._dt = Decimal(dt)
        self._bw_dt = Decimal(bw_dt)
        self._check_dt()
        self.states_sampling = states_sampling
        self.states_ts = states_ts
        self.gof_terms = gof_terms
        self.do_fc = do_fc
        self.do_fcd = do_fcd
        self.window_size = window_size
        self.window_step = window_step
        self.sim_seed = sim_seed
        self.exc_interhemispheric = exc_interhemispheric
        self.force_cpu = force_cpu
        self.force_gpu = force_gpu
        self.bold_remove_s = bold_remove_s
        self.fcd_drop_edges = fcd_drop_edges
        self.noise_segment_length = noise_segment_length
        self.noise_out = noise_out
        self.sim_verbose = sim_verbose
        self.progress_interval = progress_interval
        # get number of available hardware
        self.avail_gpus = utils.avail_gpus()
        self.avail_cpus = multiprocessing.cpu_count()
        # set number of nodes
        self.nodes = self.sc.shape[0]
        # load sc distance if provided, and set delay flag accordingly
        self.input_sc_dist = sc_dist
        if self.input_sc_dist is None:
            self.sc_dist = np.zeros_like(self.sc, dtype=float)
            self.do_delay = False
        else:
            if isinstance(self.input_sc_dist, (str, os.PathLike)):
                if self.input_sc_dist.endswith(".txt"):
                    self.sc_dist = np.loadtxt(self.input_sc_dist)
                elif self.input_sc_dist.endswith(".npy"):
                    self.sc_dist = np.load(self.input_sc_dist)
            else:
                self.sc_dist = self.input_sc_dist
            self.do_delay = True
        # set Ballon-Windkessel parameters
        self.bw_params = bw_params
        # initialze w_IE_list as all 0s if do_fic
        self.param_lists = dict([(k, None) for k in self.global_param_names + self.regional_param_names + ['v']])
        # determine output directory
        self.input_out_dir = out_dir
        if self.input_out_dir == "same":
            assert isinstance(self.input_sc, (str, os.PathLike)), (
                "When `out_dir` is set to 'same' `sc` must be"
                "a path-like string"
            )
            self.out_dir = self.input_sc.replace(".txt", "").replace(".npy", "")
        else:
            self.out_dir = self.input_out_dir
        # keep track of the last N and config to determine if force_reinit
        # is needed in the next run if any are changed
        self.last_N = 0
        self.last_config = self.get_config(for_reinit=True)
        # keep track of the iterations for iterative algorithms
        self.it = 0
        # run post-initialization steps
        self.post_init()

    def post_init(self):
        """
        Post-initilaization hook that normally does nothing, but
        can be overridden in derived classes to add custom
        post-initialization steps.
        """
        pass

    @property
    def N(self):
        return self._N

    @N.setter
    def N(self, N):
        self._N = N
        if hasattr(self, "_do_delay") and (not self.do_delay):
            self.param_lists["v"] = np.zeros(self._N, dtype=float)

    @property
    def nodes(self):
        return self._nodes
    
    @nodes.setter
    def nodes(self, nodes):
        self._nodes = nodes
        max_nodes = max_nodes_many if many_nodes_flag else max_nodes_reg
        # determine if with this number of nodes co-launch mode will be enabled
        # in co-launch mode multiple blocks are occupied by a single (massive) simulation
        self._co_launch = (self.nodes > max_nodes) and (not self.use_cpu)
        if self._co_launch:
            if not many_nodes_flag:
                raise NotImplementedError(
                    f"With {self.nodes} nodes in the current installation of the toolbox"
                    " simulations will fail. Please reinstall the package from source after"
                    " `export CUBNM_MANY_NODES=1`")
            if self.do_fcd:
                raise NotImplementedError(
                    "With many nodes, FCD calculation is not supported."
                    " Set do_fcd to False."
                )
            if self.dt < 1.0:
                logger.warning(
                    "When runnig simulations with large number "
                    "of nodes it is recommended to set the model dt to "
                    ">=1.0 msec for better performance."
                )

    @property
    def duration(self):
        return self._duration
    
    @duration.setter
    def duration(self, duration):
        self._duration = duration
        duration_decimal = Decimal(duration)
        duration_msec_decimal = duration_decimal * Decimal(1000)
        if duration_msec_decimal % 1 != 0:
            raise ValueError(
                f"Duration {duration} seconds equals {duration_msec_decimal} milliseconds, "
                "which is not a whole number. Duration must be expressible in whole milliseconds, "
                "i.e., must not have more than 3 decimal places."
            )
        self.duration_msec = int(duration_msec_decimal)
        if hasattr(self, "_noise_segment_length") and (self.noise_segment_length is None):
            self.noise_segment_length = self.duration

    @property
    def TR(self):
        return self._TR

    @TR.setter
    def TR(self, TR):
        self._TR = TR
        self.TR_msec = int(TR * 1000)
        if hasattr(self, "_states_sampling") and ((self.states_sampling is None) | (self.states_sampling == self.TR)):
            self.states_sampling = self.TR
        if hasattr(self, "_window_size"):
            self.window_size_TRs = int(np.round(self.window_size / (self.TR*2))) * 2
        if hasattr(self, "_window_step"):
            self.window_step_TRs = int(np.round(self.window_step / self.TR))

    @property
    def states_sampling(self):
        return self._states_sampling

    @states_sampling.setter
    def states_sampling(self, states_sampling):
        # raise an error if sampling rate is below model's outer loop dt
        if (states_sampling is not None) and (states_sampling < 0.001):
            raise ValueError(
                "states_sampling cannot be lower than 0.001s "
                "which is model dt"
            )
        if hasattr(self, "_TR") and (states_sampling is None):
            # also set it to TR when None
            self._states_sampling = self.TR
        else:
            self._states_sampling = states_sampling
        # convert states_sampling to msec
        self.states_sampling_msec = int(self.states_sampling * 1000)

    @property
    def n_states_samples_remove(self):
        return int(self.bold_remove_s / self.states_sampling)

    @property
    def noise_segment_length(self):
        return self._noise_segment_length

    @noise_segment_length.setter
    def noise_segment_length(self, noise_segment_length):
        if hasattr(self, "_duration") and (noise_segment_length is None):
            self._noise_segment_length = self.duration
        else:
            self._noise_segment_length = noise_segment_length

    @property
    def dt(self):
        return self._dt
    
    @dt.setter
    def dt(self, dt):
        self._dt = Decimal(dt)
        self._check_dt()

    @property
    def bw_dt(self):
        return self._bw_dt
    
    @bw_dt.setter
    def bw_dt(self, bw_dt):
        self._bw_dt = Decimal(bw_dt)
        self._check_dt()

    @property
    def bw_it(self):
        """
        Number of Balloon-Windkessel integration steps (outer loop)
        """
        return int(self.duration_msec / self.bw_dt)

    @property
    def inner_it(self):
        """
        Number of model integration steps (inner loop) within 
        a Balloon-Windkessel integration step
        """
        return int(self.bw_dt / self.dt)

    @property
    def bw_params(self):
        return self._bw_params

    @bw_params.setter
    def bw_params(self, bw_params):
        self._bw_params = bw_params
        params = utils.get_bw_params(self._bw_params)
        set_const("k1", params["k1"])
        set_const("k2", params["k2"])
        set_const("k3", params["k3"])

    @property
    def do_delay(self):
        return self._do_delay

    @do_delay.setter
    def do_delay(self, do_delay):
        self._do_delay = do_delay
        if hasattr(self, "_dt") and self.do_delay and (self.dt < 1.0):
            logger.warning(
                "When using delay in the simulations "
                "it is recommended to set the model dt to >= "
                "1.0 msec for better performance."
            )

    @property
    def window_size(self):
        return self._window_size

    @window_size.setter
    def window_size(self, window_size):
        self._window_size = window_size
        # calculate nearest even number of TRs
        # TODO: enable odd window sizes
        self.window_size_TRs = int(np.round(window_size / (self.TR*2))) * 2

    @property
    def window_step(self):
        return self._window_step
    
    @window_step.setter
    def window_step(self, window_step):
        self._window_step = window_step
        self.window_step_TRs = int(np.round(window_step / self.TR))

    @property
    def do_fc(self):
        return self._do_fc

    @do_fc.setter
    def do_fc(self, do_fc):
        if hasattr(self, "gof_terms") and (not do_fc):
            if ('+fc_corr' in self.gof_terms) or ('-fc_diff' in self.gof_terms) or ('-fc_normec' in self.gof_terms):
                raise ValueError("Cannot calculate FC goodness-of-fit terms without FC."
                                 " Set do_fc to True or remove FC-related goodness-of-fit"
                                 " terms.")
        if getattr(self, "_do_fcd", False) and (not do_fc):
            raise ValueError("Cannot calculate FCD without FC. Set do_fcd to False.")
        self._do_fc = do_fc

    @property
    def do_fcd(self):
        return self._do_fcd

    @do_fcd.setter
    def do_fcd(self, do_fcd):
        if hasattr(self, "_do_fc") and self.do_fc:
            self._do_fcd = do_fcd
        elif do_fcd:
            raise ValueError("Cannot calculate FCD without FC")
        else:
            self._do_fcd = False
        if hasattr(self, "gof_terms") and (not self._do_fcd) and ('-fcd_ks' in self.gof_terms):
            raise ValueError("Cannot calculate FCD goodness-of-fit terms without FCD."
                                " Set do_fcd to True or remove FCD-related goodness-of-fit"
                                " terms.")

    @property
    def labels(self):
        """
        Labels of parameters and state variables
        to use in plots and reports
        """
        return {}

    def _check_dt(self):
        """
        Check if integrations steps are valid
        """
        assert self.bw_dt >= self.dt, (
            "Ballon-Windkessel dt must be greater than or equal to model dt"
        )
        assert (self.bw_dt % self.dt) == Decimal(0), (
            "Ballon-Windkessel dt must be divisible by model dt"
        )
        assert (self.duration_msec % self.bw_dt) == Decimal(0), (
            "Duration must be divisible by Ballon-Windkessel dt"
        )

    def _set_default_params(self, missing=True):
        """
        Set default parameters for the simulations.
        
        Parameters
        ----------
        missing: :obj:`bool`, optional
            If True (default), only sets parameters that are currently None
            or have an incorrect length in self.param_lists. 
            If False, always overwrites all listed defaults.
        """
        defaults = copy.copy(self.default_params)
        if self.do_delay and ("v" not in defaults):
            defaults["v"] = 0.5
        for param, v in defaults.items():
            if (
                (not missing) or
                (self.param_lists.get(param) is None) or 
                (self.param_lists.get(param, np.array([0])).shape[0] != self.N)
                # the last condition is to ensure that in subsequent generations
                # of evolutionary algorithm the default regional parameters
            ):
                if v is None:
                    raise ValueError(
                        f"Parameter {param} does not have a default value"
                        " and must be set as a fixed or free parameter"
                    )
                if param == "v" and self.do_delay:
                    # v = 0.5 is used to push the simulations to a difficult
                    # case (with prolonged history to be stored) to make sure
                    # everything works in extreme cases, but biologically this
                    # is a low conduction velocity
                    logger.warning(
                        "The default conduction velocity (v) is set to 0.5 m/s. "
                        "Which may be biologically not realistic (too low). "
                        "Consider setting it as a free parameter or use a higher fixed value."
                    )
                if param in self.global_param_names:
                    self.param_lists[param] = np.repeat(v, self.N)
                else:
                    if (param == "wIE") and self.do_fic:
                        continue
                    self.param_lists[param] = np.full((self.N, self.nodes), v)

    def get_config(self, include_N=False, for_reinit=False):
        """
        Get the configuration of the simulation group

        Parameters
        ----------
        include_N: :obj:`bool`
            include N in the output config
            is ignored when for_reinit is True
        for_reinit: :obj:`bool`
            include the parameters that need reinitialization of the
            simulation core session if changed

        Returns
        -------
        config: :obj:`dict`
            dictionary of simulation group configuration
        """
        config = {
            "duration": self.duration,
            "TR": self.TR,
            "nodes": self.nodes,
            "states_ts": self.states_ts,
            "states_sampling": self.states_sampling,
            "do_fc": self.do_fc,
            "do_fcd": self.do_fcd,
            "window_size": self.window_size,
            "window_step": self.window_step,
            "window_size_TRs": self.window_size_TRs,
            "window_step_TRs": self.window_step_TRs,
            "sim_seed": self.sim_seed,
            "exc_interhemispheric": self.exc_interhemispheric,
            "force_cpu": self.force_cpu,
            "force_gpu": self.force_gpu,
            "hardware_used": "cpu" if self.use_cpu else "gpu",
            "bold_remove_s": self.bold_remove_s,
            "fcd_drop_edges": self.fcd_drop_edges,
            "noise_segment_length": self.noise_segment_length,
        }
        if not for_reinit:
            # these are features that if changed
            # will not require reinitialization in core
            # (reallocation of memory and/or recalculation
            # of noise)
            config["out_dir"] = self.input_out_dir
            config["gof_terms"] = self.gof_terms
            config["sc"] = self.input_sc
            config["sc_dist"] = self.input_sc_dist
            config["bw_params"] = self.bw_params
            config["version"] = __version__
            if include_N:
                config["N"] = self.N
        return config
    
    @property
    def _model_config(self):
        """
        Internal model configuration used in the simulation

        Returns
        -------
        model_config: :obj:`dict`
            Dictionary of internal model configurations
        """
        model_config = {
            'do_fc': str(int(self.do_fc)),
            'do_fcd': str(int(self.do_fcd)),
            'bold_remove_s': str(self.bold_remove_s),
            'exc_interhemispheric': str(int(self.exc_interhemispheric)),
            'window_size': str(self.window_size_TRs),
            'window_step': str(self.window_step_TRs),
            'drop_edges': str(int(self.fcd_drop_edges)),
            'noise_time_steps': str(int(self.noise_segment_length*1000)), 
            'verbose': str(int(self.sim_verbose)),
            'progress_interval': str(int(self.progress_interval)),
        }
        return model_config

    @property
    def use_cpu(self):
        """
        Check if CPU should be used for the simulations
        based on the available hardware, compilation options, 
        and user settings
        """
        return (
            (self.force_cpu | 
            (not gpu_enabled_flag) | 
            (self.avail_gpus == 0)) & 
            (not self.force_gpu)
        )

    def run(self, force_reinit=False):
        """
        Run the simulations in parallel (as possible) on GPU/CPU
        through the :func:`cubnm._core.run_simulations` function which runs
        compiled C++/CUDA code.

        Parameters
        ----------
        force_reinit: :obj:`bool`
            force reinitialization of the session.
            At the beginning of each session (when `cubnm` is imported)
            some variables are initialized on CPU/GPU and reused in
            every run. Set this to True if you want to reinitialize
            these variables. This is rarely needed.
        """
        # TODO: add assertions to make sure all the input data is complete
        # check if reinitialization is needed (user request, change of N,
        # of change of other config)
        curr_config = self.get_config(for_reinit=True)
        force_reinit = (
            force_reinit
            | (self.N != self.last_N)
            | (curr_config != self.last_config)
        )
        # check if running on Jupyter and print warning that
        # simulations are running but progress will not be shown
        if utils.is_jupyter():
            logger.warning(
                "Progress cannot be shown in real time "
                "when using Jupyter, but simulations are running "
                "in background. Please wait..."
            )
        # convert self.param_lists to flattened and contiguous arrays of global
        # and local parameters
        global_params_arrays = []
        regional_params_arrays = []
        for param in self.global_param_names:
            global_params_arrays.append(np.ascontiguousarray(self.param_lists[param])[np.newaxis, :])
        for param in self.regional_param_names:
            regional_params_arrays.append(np.ascontiguousarray(self.param_lists[param].flatten()))
        self._global_params = np.vstack(global_params_arrays)
        self._regional_params = np.vstack(regional_params_arrays)
        # specify fixed or variable SCs
        if ((self.sc.ndim == 2) or (self.sc.shape[0] == 1)):
            sc = np.ascontiguousarray(self.sc.flatten())[None, :]
            sc_indices = np.repeat(0, self.N).astype(np.intc)
        else:
            sc = np.ascontiguousarray(self.sc.reshape(self.sc.shape[0], -1))
            sc_indices = np.ascontiguousarray(self.sc_indices.astype(np.intc))
        # make sure sc indices are correctly in {0, ..., N_SCs-1}
        assert sc.shape[0] == np.unique(sc_indices).size
        assert (np.sort(np.unique(sc_indices)) == np.arange(np.unique(sc_indices).size)).all()
        assert sc_indices.size == self.N
        # run the simulations
        args = (
            self.model_name,
            sc,
            sc_indices,
            np.ascontiguousarray(self.sc_dist.flatten()),
            self._global_params,
            self._regional_params,
            np.ascontiguousarray(self.param_lists["v"]),
            self._model_config,
            True, # ext_out: a depracated option that is always True now
            self.states_ts,
            self.noise_out,
            self.do_delay,
            force_reinit,
            self.use_cpu,
            self.N,
            self.nodes,
            self.duration_msec,
            self.TR_msec,
            self.states_sampling_msec,
            self.sim_seed,
            float(self.dt),
            float(self.bw_dt),
        )
        try:
            out = run_simulations(*args)
        except RuntimeError as e:
            if "CUDA Error" in str(e):
                hint = (
                    "CUDA error occurred. This might be due to the data exceeding "
                    "the available GPU memory."
                )
                raise RuntimeError(hint) from e
        # avoid reinitializing GPU in the next runs
        # of the same group
        self.last_N = self.N
        self.last_config = curr_config
        self.it += 1
        # process output
        self._process_out(out)

    def _process_out(self, out):
        """
        Assigns model outputs (as arrays) to object attributes
        with correct shapes, names and types.

        Parameters
        ----------
        out: :obj:`tuple`
            output of ``cubnm._core.run_simulations`` function

        Notes
        -----
        The simulation outputs are assigned to the following object attributes:

        - init_time: :obj:`float`
            initialization time of the simulations
        - run_time: :obj:`float`
            run time of the simulations
        - sim_bold: :obj:`np.ndarray`
            simulated BOLD time series. Shape: (N_SIMS, duration/TR, nodes)
        - sim_states: :obj:`dict` of :obj:`np.ndarray`
            simulated state variables with keys as state names
            and values as arrays with the shape (N_SIMS, nodes)
            when ``states_ts`` is False, and (N_SIMS, duration/TR, nodes)
            when ``states_ts`` is True

        If ``do_fc`` is ``True``, additionally includes:

        - sim_fc_trils: :obj:`np.ndarray`
            simulated FC lower triangle. Shape: (N_SIMS, n_pairs)

        If ``do_fcd`` is ``True``, additionally includes:

        - sim_fcd_trils: :obj:`np.ndarray`
            simulated FCD lower triangle. Shape: (N_SIMS, n_window_pairs)
        """
        # assign the output to object attributes
        self.__dict__.update(out)
        # reshape the simulated BOLD, FC and FCD to (N, ...)
        self.sim_bold = self.sim_bold.reshape(self.N, -1, self.nodes)
        # process sim states into a dictionary
        # TODO: ensure each state's array is a view and not a copy of the original
        self.sim_states = {}
        for state_i, state_name in enumerate(self.state_names):
            self.sim_states[state_name] = self._sim_states[state_i, :, :]
            if self.states_ts:
                self.sim_states[state_name] = self.sim_states[state_name].reshape(self.N, -1, self.nodes)
        if self.do_fc:
            self.sim_fc_trils = self.sim_fc_trils.reshape(self.N, -1)
            if self.do_fcd:
                self.sim_fcd_trils = self.sim_fcd_trils.reshape(self.N, -1)

    def get_state_averages(self):
        """
        Get the averages of state variables across time and nodes
        for each simulation.

        Returns
        -------
        state_averages: :obj:`pd.DataFrame`
            DataFrame of state averages with columns as state names
            and rows as simulations
        """
        state_averages = {}
        for state_var in self.state_names:
            if self.states_ts:
                state_averages[state_var] = (
                    self.sim_states[state_var]
                    [:, self.n_states_samples_remove:, :]
                    .mean(axis=1).mean(axis=1)
                )
            else:
                state_averages[state_var] = self.sim_states[state_var].mean(axis=1)
        return pd.DataFrame(state_averages)

    def get_noise(self):
        """
        Get the (recreated) noise time series. This requires recreation
        of the noise array if noise segmenting is on. Noise will be
        recreated based on shuffling indices of nodes and time steps,
        similar to how it is done in the core code.

        Returns
        -------
        noise: :obj:`np.ndarray`
            Noise segment array. Shape: (n_noise, nodes, bw_it, inner_it)
        """
        # raise an error if noise was not saved in the simulation
        if not self.noise_out:
            raise ValueError(
                "Noise was not saved in the simulation."
                " Set `noise_out` to true and redo the simulation."
                )
        # raise an error if simulation is not run yet
        if not hasattr(self, "_noise"):
            raise ValueError(
                "Simulation must be run for the noise to be calculated"
                )
        if not noise_segment_flag:
            # just reshape the whole noise array to the correct shape
            return (
                self._noise.
                reshape(self.bw_it, self.inner_it, self.nodes, -1)
                .transpose(3, 2, 0, 1)
            )
        # otherwise recreate the noise time series otherwise
        # first reshape the noise into (noise_segment_outer_it, inner_it, nodes, n_noise)
        # get number of outer (bw) time steps per noise segment
        noise_segment_outer_it = int(int(self.noise_segment_length*1000) / self.bw_dt)
        noise_segment = self._noise.reshape(
            noise_segment_outer_it, self.inner_it, self.nodes, -1
        )
        # then shuffle the noise in each repeat based on the shuffling indices
        noise_all = []
        for r in range(self._shuffled_nodes.shape[0]):
            noise_all.append(
                noise_segment[
                    self._shuffled_ts[r] # shuffle outer time steps
                ][
                    :, :, self._shuffled_nodes[r] # shuffle nodes
                ]
            )
        # conctaenate the shuffled noise repeats
        noise_all = np.concatenate(noise_all, axis=0)
        # crop the last part of the noise that is not used
        noise_all = noise_all[:self.bw_it]
        # transpose to (n_noise, nodes, bw_it, inner_it)
        return noise_all.transpose(3, 2, 0, 1)

    def get_sim_fc(self, idx):
        """
        Get the simulated FC of a given simulation ``idx``
        as a square matrix.

        Parameters
        ----------
        idx: :obj:`int`
            index of the simulation to get the FC for

        Returns
        -------
        fc: :obj:`np.ndarray`
            simulated FC matrix of shape (nodes, nodes)
            for the simulation with index ``idx``
        """
        if not self.do_fc:
            raise ValueError("FC is not calculated in this simulation group")
        if idx >= self.N:
            raise IndexError("Index out of range")
        # get the FC lower triangle
        fc_tril = self.sim_fc_trils[idx].copy()
        # convert it to a square matrix
        sim_fc = np.zeros((self.nodes, self.nodes), dtype=float)
        if self.exc_interhemispheric:
            # when interhemispheric connections are excluded
            # the FC matrix is split into two halves
            # and the lower triangle is filled in each half
            half_nodes = self.nodes // 2
            sim_fc[:half_nodes, :half_nodes][
                np.tril_indices(half_nodes,-1)
            ] = fc_tril[:fc_tril.shape[0] // 2]
            sim_fc[half_nodes:, half_nodes:][
                np.tril_indices(half_nodes,-1)
            ] = fc_tril[fc_tril.shape[0] // 2:]
            # set the rest to NaNs
            sim_fc[:half_nodes, half_nodes:] = np.NaN
            sim_fc[half_nodes:, :half_nodes] = np.NaN
        else:
            # fill the lower triangle of the square matrix
            sim_fc[np.tril_indices(self.nodes,-1)] = fc_tril
        # make the matrix symmetric
        sim_fc += sim_fc.T
        # fill the diagonal with 1s
        np.fill_diagonal(sim_fc, 1.0)
        return sim_fc

    def get_sim_fcd(self, idx):
        """
        Get the simulated FCD of a given simulation ``idx``
        as a square matrix.

        Parameters
        ----------
        idx: :obj:`int`
            index of the simulation to get the FC for
        
        Returns
        -------
        fcd: :obj:`np.ndarray`
            simulated FC matrix of shape (n_windows, n_windows)
            for the simulation with index ``idx``
        """
        if not self.do_fcd:
            raise ValueError("FCD is not calculated in this simulation group")
        if idx >= self.N:
            raise IndexError("Index out of range")
        # get the FCD lower triangle
        fcd_tril = self.sim_fcd_trils[idx].copy()
        n_pairs = fcd_tril.shape[0]
        # determine number of windows
        # this is not returned from the core
        # and its direct calculation without using
        # a loop isn't trivial (because of different
        # conditions etc.; or at least I don't know
        # how to do it).
        # therefore, we use the (known) number of pairs
        # and solve for n_pairs = n_windows * (n_windows - 1) / 2
        # aka:
        n_windows = (1 + np.sqrt(1 + 8 * n_pairs)) / 2
        assert n_windows.is_integer() # just a sanity check
        n_windows = int(n_windows)
        # convert it to a square matrix
        sim_fcd = np.zeros((n_windows, n_windows), dtype=float)
        # fill the lower triangle of the square matrix
        sim_fcd[np.tril_indices(n_windows,-1)] = fcd_tril
        # make the matrix symmetric
        sim_fcd += sim_fcd.T
        # fill the diagonal with 1s
        np.fill_diagonal(sim_fcd, 1.0)
        return sim_fcd

    def slice(self, key, inplace=False):
        """
        Slice the simulation group to a single simulation

        Parameters
        ----------
        key: :obj:`int`
            index of the simulation to slice
        inplace: :obj:`bool`
            the object will be sliced in place
            and therefore the data of other simulations
            will be removed. Otherwise a new object
            copied from the current object will be returned.
        
        Returns
        -------
        obj: :obj:`cubnm.sim.SimGroup`
            sliced simulation group
        """
        if not isinstance(key, int):
            raise ValueError("Only integer indexing is supported")
        if key >= self.N:
            raise IndexError("Index out of range")
        if inplace:
            obj = self
        else:
            # create a (shallow) copy
            # and not a deep copy as the
            # data may be very large
            obj = copy.copy(self)
        obj.param_lists = {k: v[key][np.newaxis, ...] for k, v in self.param_lists.items()}
        obj.sim_bold = self.sim_bold[key][np.newaxis, ...]
        obj.sim_states = {k: v[key][np.newaxis, ...] for k, v in self.sim_states.items()}
        if obj.do_fc:
            obj.sim_fc_trils = self.sim_fc_trils[key][np.newaxis, ...]
            if obj.do_fcd:
                obj.sim_fcd_trils = self.sim_fcd_trils[key][np.newaxis, ...]
        obj.N = 1
        return obj

    @property
    def _to_clear(self):
        """
        List of attributes to be cleared when clear() is called
        """
        attrs = [
            "sim_bold", 
            "sim_fc_trils", 
            "sim_fcd_trils", 
            "sim_states",
            "_sim_states",
            "_noise",
            "_shuffled_nodes",
            "_shuffled_ts",
        ]
        return [attr for attr in attrs if hasattr(self, attr)]

    def clear(self):
        """
        Clear the simulation outputs
        """
        for attr in self._to_clear:
            if hasattr(self, attr):
                delattr(self, attr)
        gc.collect()

    def _problem_init(self, problem):
        """
        Extends BNMProblem initialization if needed.
        By default it doesn't do anything.

        Parameters
        ----------
        problem: :obj:`cubnm.optimize.BNMProblem`
            optimization problem object
        """
        pass

    def _problem_evaluate(self, problem, X, out, *args, **kwargs):
        """
        Extends BNMProblem evaluation if needed.
        By default it doesn't do anything.

        Parameters
        ----------
        X: :obj:`np.ndarray`
            the normalized parameters of current population in range [0, 1]. 
            Shape: (N, ndim)
        out: :obj:`dict`
            the output dictionary to store the results with keys 'F' and 'G'.
            Currently only 'F' (cost) is used.
        *args, **kwargs
        """
        pass

    def score(
            self, 
            emp_fc_tril=None, 
            emp_fcd_tril=None, 
            emp_bold=None, 
            force_cpu=False,
            usable_mem=None
        ):
        """
        Calcualates individual goodness-of-fit terms and aggregates them.

        .. note::
            If `emp_bold` is provided, `emp_fc_tril` and `emp_fcd_tril` will be ignored.

        .. note::
            For each measure, if the value is NaN, it will be set to the "worst" possible value.
            NaNs may occur in simulated FCD or FC. For example, in the rWWEx model, when excitation
            is too high and noise is low, `S` and in turn `BOLD` in some areas may become saturated
            and show no variability. This can result in correlations of their BOLD signals with
            other nodes (within certain dynamic windows) being NaN.

        Parameters
        --------
        emp_fc_tril: :obj:`np.ndarray` or None
            1D array of empirical FC lower triangle. Shape: (edges,)
        emp_fcd_tril: :obj:`np.ndarray` or None
            1D array of empirical FCD lower triangle. Shape: (window_pairs,)
        emp_bold: :obj:`np.ndarray` or None
            cleaned and parcellated empirical BOLD time series. Shape: (nodes, volumes)
            Motion outliers should either be excluded (not recommended as it disrupts
            the temporal structure) or replaced with zeros.
            If provided emp_fc_tril and emp_fcd_tril will be ignored.
        force_cpu: :obj:`bool`
            force CPU for the calculations. Otherwise if GPU is available
            some of the scores (including "+fc_corr", "-fcd_ks") 
            will be calculated on GPU.
        usable_mem: :obj:`int`
            amount of available GPU memory to be used in bytes.
            If None, 80% of the free memory will be used.

        Returns
        -------
        scores: :obj:`pd.DataFrame`
            The goodness of fit measures (columns) of each simulation (rows)
        """
        # calculate empirical FC and FCD if BOLD is provided
        if emp_bold is not None:
            if (emp_fc_tril is not None) or (emp_fcd_tril is not None):
                logger.warning(
                    "Both empirical BOLD and empirical FC/FCD are "
                    "provided. Empirical FC/FCD will be calculated based on "
                    "BOLD and will be overwritten."
                )
            if self.do_fc:
                emp_fc_tril = utils.calculate_fc(emp_bold, self.exc_interhemispheric, return_tril=True)
            if self.do_fcd:
                emp_fcd_tril = utils.calculate_fcd(
                    emp_bold, 
                    self.window_size_TRs, 
                    self.window_step_TRs, 
                    drop_edges=self.fcd_drop_edges,
                    exc_interhemispheric=self.exc_interhemispheric,
                    return_tril=True,
                    return_dfc = False
                )
        # + => aim to maximize; - => aim to minimize
        # get list of columns to be calculated
        # based on availability of FC and FCD (simulated and empirical)
        # and gof_terms
        columns = []
        if self.do_fc and (emp_fc_tril is not None):
            columns += list(set(["+fc_corr", "-fc_diff", "-fc_normec"]) & set(self.gof_terms))
            assert emp_fc_tril.size == self.sim_fc_trils.shape[1], (
                "Empirical FC lower triangle size does not match the simulated FC size"
            )
        if self.do_fcd and (emp_fcd_tril is not None):
            columns += list(set(["-fcd_ks"]) & set(self.gof_terms))
        columns += ["+gof"]
        scores = pd.DataFrame(index=range(self.N), columns=columns, dtype=float)
        # vectorized calculation of some measures on CPU
        for column in columns:
            if column == "-fc_diff":
                scores.loc[:, column] = -np.abs(
                    self.sim_fc_trils.mean(axis=1) - emp_fc_tril.mean()
                )
            elif column == "-fc_normec":
                scores.loc[:, column] = -np.linalg.norm(
                    self.sim_fc_trils - emp_fc_tril[None, :],
                    axis=1,
                ) / (2 * np.sqrt(emp_fc_tril.size))
        # calculation of fc_corr and fcd_ks per simulation on CPU
        # or in parallel batches on GPU (if available and CuPy is installled)
        if (self.use_cpu or force_cpu or (not has_cupy)):
            # TODO: run in parallel on CPU cores when
            # self.avail_cpus > 1
            for idx in range(self.N):
                for column in columns:
                    if column == "+fc_corr":                    
                        # TODO: vectorize FC corr calculation
                        scores.loc[idx, column] = scipy.stats.pearsonr(
                            self.sim_fc_trils[idx], emp_fc_tril
                        ).statistic
                    elif column == "-fcd_ks":
                        scores.loc[idx, column] = -scipy.stats.ks_2samp(
                            self.sim_fcd_trils[idx], emp_fcd_tril
                        ).statistic
        else:
            # calculate on GPU
            if usable_mem is None:
                # get amount of available GPU memory as
                # 80% of the free memory
                free_mem, _ = cp.cuda.runtime.memGetInfo()
                usable_mem = int(free_mem * 0.8)
            for column in columns:
                if column == "+fc_corr":
                    scores.loc[:, column] = utils.fc_corr_device(self.sim_fc_trils, emp_fc_tril, usable_mem)
                elif column == "-fcd_ks":
                    scores.loc[:, column] = -utils.fcd_ks_device(self.sim_fcd_trils, emp_fcd_tril, usable_mem)
        # fill NaNs with the worst possible value
        for column in columns:
            scores.loc[:, column] = scores.loc[:, column].fillna(WORST_SCORES.get(column, np.nan))
        # combined the selected terms into gof (that should be maximized)
        scores.loc[:, "+gof"] = scores.loc[:, self.gof_terms].sum(axis=1)
        return scores
    
    def _get_save_data(self):
        """
        Get the simulation outputs and parameters to be saved to disk

        Returns
        -------
        out_data: :obj:`dict`
            dictionary of simulation outputs and parameters
        """
        out_data = dict(
            sim_bold=self.sim_bold,
        )
        out_data.update(self.sim_states)
        if self.do_fc:
            out_data['sim_fc_trils'] = self.sim_fc_trils
        if self.do_fcd:
            out_data['sim_fcd_trils'] = self.sim_fcd_trils
        out_data.update(self.param_lists)
        return out_data

    def save(self):
        """
        Save simulation outputs to disk as an npz file.
        """
        assert self.out_dir is not None, (
            "Cannot save the simulations when `.out_dir`"
            "is not set"
        )
        os.makedirs(self.out_dir, exist_ok=True)
        out_data = self._get_save_data()
        np.savez_compressed(os.path.join(self.out_dir, "sim_data.npz"), **out_data)

    @classmethod
    def _get_test_configs(cls, cpu_gpu_identity=False):
        """
        Get configs for testing the simulations

        Parameters
        ----------
        cpu_gpu_identity: :obj:`bool`
            indicates whether configs are for CPU/GPU identity tests
            in which case force_cpu is not included in the configs
            since tests will be done on both CPU and GPU

        Returns
        -------
        configs: :obj:`dict` of :obj:`list`
        """
        configs = {}
        # for compatibality with previously-stored
        # expected simulation files force_cpu must
        # be the first item
        # TODO: fix this and recompute the expected files
        if not cpu_gpu_identity:
            configs['force_cpu'] = [0, 1]
        configs['do_delay'] = [0, 1]
        return configs

    @classmethod
    def _get_test_instance(cls, opts):
        """
        Initializes an instance that is used in tests

        Parameters
        ----------
        opts: :obj:`dict`
            dictionary of test options

        Returns
        -------
        sim_group: :obj:`cubnm.sim.SimGroup`
            simulation group object of the test simulation
            which is not run yet
        """
        do_delay = bool(opts.pop('do_delay'))
        if do_delay:
            sc_dist = datasets.load_sc('length', 'schaefer-100')
            dt = '1.0'
        else:
            sc_dist = None
            dt = '0.1'
        sim_group = cls(
            duration=60,
            TR=1,
            window_size=10,
            window_step=2,
            sc=datasets.load_sc('strength', 'schaefer-100'),
            sc_dist=sc_dist,
            dt = dt,
            sim_verbose=False,
            **opts
        )
        return sim_group

class MultiSimGroupMixin:
    def __init__(self, sim_groups):
        """
        Mixin for combining multiple simulation groups into one,
        which is intended for batch optimization, i.e. running multiple
        optimizations at the same time on GPU.
        """
        self.children = sim_groups
        # copy all attributes of the first child
        # it is important to deep copy the attributes as
        # some (e.g. param_lists) are modifiable
        # this also assumes that all children have the same attributes
        # except SC which can be variable
        # TODO: add a check that this is the case
        self.__dict__.update(copy.deepcopy(sim_groups[0].__dict__))

    @property
    def N(self):
        return sum([child.N for child in self.children])
    
    def run(self, **kwargs):
        """
        Runs merged simulations of all children
        by concatenating SCs and parameters
        and then running all simulations in parallel
        as a single merged simulation group.

        Parameters
        ----------
        **kwargs
            keyword arguments to be passed to `cubnm.sim.SimGroup.run` method
        """
        # concatenate SCs while allowing variable SCs
        # across children
        for child_i, child in enumerate(self.children):
            # make a copy of current child's SCs
            child_scs = child.sc.copy()
            # make scs 3D (n_scs, nodes, nodes)
            if (child_scs.ndim == 2):
                child_scs = child_scs[None, :, :]
            # create sc_indices as all zeros if only one SC is used
            # in current child
            if (child_scs.shape[0] == 1):
                child_sc_indices = np.zeros(child.N, dtype=int)
            else:
                child_sc_indices = np.array(child.sc_indices).copy()
            # initialize merged sim group SC as the first child SC(s)
            if child_i == 0:
                self.sc = child_scs
                self.sc_indices = child_sc_indices
            else:
                # for subsequent children check if SC is already in self.sc
                # (and then update the indices to point to correct sc for each
                # simulation), otherwise add the SC to self.sc (and update the
                # indices with a new index pointing to the new SC)
                sc_indices = child_sc_indices
                for sc_i, sc in enumerate(child_scs):
                    if ~np.isclose(sc, self.sc).all(axis=(1, 2)).any():
                        # if SC is not already in self.sc, add it
                        # as a new SC with a new index
                        self.sc = np.concatenate([self.sc, sc[None, :, :]], axis=0)
                        new_sc_idx = self.sc_indices.max()+1
                        sc_indices[sc_indices == sc_i] = new_sc_idx
                    else:
                        # if SC is already in self.sc, find its index
                        # and just update the sc_indices to reflect the duplicate
                        # which already exist. In this case there is no need to
                        # add the SC to self.sc as it already exists there
                        dupl_sc_idx = np.where(np.isclose(sc, self.sc).all(axis=(1, 2)))[0][0]
                        sc_indices[sc_indices == sc_i] = dupl_sc_idx
                    # update self.sc_indices
                    self.sc_indices = np.concatenate([self.sc_indices, sc_indices], axis=0)
        # convert sc back to 2D if same SC is used in all simulations
        self.sc = np.squeeze(self.sc)
        # concatenate parameters
        for param in self.param_lists:
            self.param_lists[param] = np.concatenate(
                [child.param_lists[param] for child in self.children]
            )
        # run the simulations
        super().run(**kwargs)

    def _process_out(self, out):
        """
        Divides simulations outputs across individual
        children SimGroup objects which will respectively
        convert the output to attributes with correct shapes,
        names and types. See :func:`cubnm.sim.base.SimGroup._process_out`
        for details.

        Parameters
        ----------
        out: :obj:`tuple`
            output of `run_simulations` function
        """
        start_idx = 0
        for child in self.children:
            end_idx = start_idx + child.N
            # break down self._regional_params and self._global_params
            # (which may be modified in run) into children as they may
            # be used in _process_out of children (e.g. in rWW)
            child._regional_params = self._regional_params[:, start_idx:end_idx]
            child._global_params = self._global_params[:, start_idx:end_idx]
            # break down simulation output elements
            out_child = {}
            for k in [
                "init_time",
                "run_time",
                "sim_bold", 
                "sim_fc_trils", 
                "sim_fcd_trils", 
                "_sim_states",
                "_global_bools",
                "_global_ints",
                "_noise",
                "_shuffled_nodes",
                "_shuffled_ts",
            ]:
                if k not in out:
                    continue
                if k in ["init_time", "run_time", "_noise", "_shuffled_nodes", "_shuffled_ts"]:
                    # shared between all simulations
                    out_child[k] = out[k]
                elif k in ["sim_bold", "sim_fc_trils", "sim_fcd_trils"]:
                    # simulation-specific outputs with shape (N, ...)
                    out_child[k] = out[k][start_idx:end_idx]
                else:
                    # simulation-specific outputs with shape (..., N, ...)
                    # TODO: it'll be easier to do this if all
                    # simulation-specific outputs have N_SIMS as
                    # the first axis
                    out_child[k] = out[k][:, start_idx:end_idx]
            # update start_idx for the next child
            start_idx += child.N
            # process the output of the child
            child._process_out(out_child)

def create_multi_sim_group(sim_group_cls):
    """
    Dynamically creates a MultiSimGroup class by combining
    a model's specific ``<Model>SimGroup`` class with 
    :class:`cubnm.sim.base.MultiSimGroupMixin`,
    which can be used in batch optimization.

    Parameters
    ----------
    sim_group_cls: :obj:`type`
        :class:`cubnm.sim.base.SimGroup` subclass, e.g. :class:`cubnm.sim.rww.rWWSimGroup`
    
    Returns
    -------
    MultiSimGroup: :obj:`type`
        :class:`MultiSimGroup` class that can be used in batch optimization
    """
    MultiSimGroup = type(
        'MultiSimGroup',
        (MultiSimGroupMixin, sim_group_cls),
        {}
    )
    return MultiSimGroup


# Autogenerated based on 'src/cubnm/sim/__init__.py.mako'
# and list of available models.
# Do not modify this autogenerated code. Instead modify template.
from .base import SimGroup, create_multi_sim_group

from .jr import JRSimGroup
from .kuramoto import KuramotoSimGroup
from .rww import rWWSimGroup
from .rwwex import rWWExSimGroup
from .wc import WCSimGroup

__all__ = [
    'SimGroup',
    'JRSimGroup',
    'KuramotoSimGroup',
    'rWWSimGroup',
    'rWWExSimGroup',
    'WCSimGroup',
]


from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from distutils.spawn import find_executable
import subprocess
import os
import fnmatch
import sys
import numpy as np
import versioneer

ROOT = os.path.abspath(os.path.dirname(__file__))

def find_file(file_name, search_path):
    result = []
    for root, dirs, files in os.walk(search_path):
        for name in fnmatch.filter(files, file_name):
            result.append(os.path.join(root, name))
    return result

def get_gpu_model(method='compiler'):
    """
    Detects the GPU type (if any) to compile for.
    It first checks for Nvidia compiler/smi and then
    ROCm.

    Parameters
    ----------
    method: {'compiler', 'smi'}
    
    Returns
    -------
    {False, 'nvidia', 'rocm'}
    """
    if method == 'smi':
        # nvidia
        try:
            output = subprocess.check_output(['nvidia-smi', '-L'])
            output = output.decode('utf-8').strip()
            gpu_list = output.split('\n')
            gpu_count = len(gpu_list)
            if gpu_count>0:
                return 'nvidia'
        except:
            pass
        # rocm
        try:
            output = subprocess.check_output(['rocm-smi', '-i', '--json'])
            output = output.decode('utf-8').strip()
            gpu_dict = json.loads(output)
            gpu_count = len(gpu_dict)
            if gpu_count>0:
                return 'rocm'
        except:
            pass
        return False
    elif method == 'compiler':
        # can be useful for compiling code on a
        # non-GPU system for running it later
        # on GPUs
        # nvidia
        if find_executable('nvcc'):
            return 'nvidia'
        elif find_executable('hipcc'):
            return 'rocm'
        else:
            return False

# specify installation options
many_nodes = not ("CUBNM_NO_MANY_NODES" in os.environ)
max_nodes_reg = os.environ.get("CUBNM_MAX_NODES_REG", 200)
max_nodes_many = os.environ.get("CUBNM_MAX_NODES_MANY", 1024)
gpu_model = get_gpu_model(method='compiler')
gpu_enabled = (gpu_model != False)
omp_enabled = not ("CUBNM_NO_OMP" in os.environ)
noise_segment = not ("CUBNM_NOISE_WHOLE" in os.environ)

# Write the flags to a temporary _flags.py file
with open(os.path.join(ROOT, "src", "cubnm", "_setup_opts.py"), "w") as flag_file:
    flag_file.write(
        "\n".join(
            [f"many_nodes_flag = {many_nodes}", 
             f"gpu_enabled_flag = {gpu_enabled}",
             f"omp_enabled_flag = {omp_enabled}",
             f"noise_segment_flag = {noise_segment}",
             f"max_nodes_reg = {max_nodes_reg}",
             f"max_nodes_many = {max_nodes_many}",
             f"gpu_model_flag = '{gpu_model}'"
            ]
        )
    )

# determine libraries shared between GPU and CPU versions
libraries = ["m", "rt"]
if omp_enabled:
    libraries.append("gomp")

_CC = os.environ.get("CC", "gcc") # this will be used for compiling GSL + restoring os $CC and $CXX at the end
_CXX = os.environ.get("CXX", "g++")
os.environ["CC"] = "g++"
os.environ["CXX"] = "g++"

# create lists of include directories
shared_includes = [
    os.path.join(ROOT,"include"),
    os.path.join(ROOT, "src", "ext"),
    np.get_include(),
    "/opt/miniforge/include" # added for conda-based cibuildwheel
]
if gpu_model == 'nvidia':
    gpu_includes = [
        "/usr/lib/cuda/include",
        "/usr/include/cuda",
    ]
elif gpu_model == 'rocm':
    ROCM_PATH = subprocess.check_output(['hipconfig', '--path']).decode('utf-8').strip()
    gpu_includes = [os.path.join(ROCM_PATH,"include", "hip")]
else:
    gpu_includes = []
all_includes = shared_includes + gpu_includes

# extra compile args shared between CPU and GPU
extra_compile_args = [
    "-std=c++11",
    "-O3",
    "-m64",
    f"-D MAX_NODES_REG={max_nodes_reg}",
    f"-D MAX_NODES_MANY={max_nodes_many}"
]
if noise_segment:
    extra_compile_args.append("-D NOISE_SEGMENT")
if many_nodes:
    extra_compile_args.append("-D MANY_NODES")
if gpu_enabled:
    extra_compile_args.append("-D GPU_ENABLED")
    if gpu_model == 'rocm':
        extra_compile_args.append("-D ENABLE_HIP")

gpu_extra_compile_args = extra_compile_args.copy()
cpu_extra_compile_args = extra_compile_args.copy()

if omp_enabled:
    cpu_extra_compile_args += [
        "-fopenmp",
        "-D OMP_ENABLED"
    ]


if gpu_enabled:
    print(f"Compiling for GPU ({gpu_model}) + CPU")
    libraries += ["bnm"]
    if gpu_model == 'nvidia':
        libraries += ["cudart_static"]
        library_dirs = [
            "/usr/lib/cuda", 
            "/usr/local/cuda/lib64",
        ]
    elif gpu_model == 'rocm':
        # TODO: is there an alternative to `cudart_static`?
        libraries += ['amdhip64']
        library_dirs = [
            os.path.join(ROCM_PATH, 'lib'),
        ]
    bnm_ext = Extension(
        "cubnm.core",
        [os.path.join("src", "ext", "core.cpp")],
        language="c++",
        extra_compile_args=cpu_extra_compile_args,
        libraries=libraries,
        include_dirs=all_includes,
        library_dirs=library_dirs + \
            [
                os.path.join(ROOT, "src", "ext"),
                "/opt/miniforge/lib" # added for conda-based cibuildwheel
            ] + \
            os.environ.get("LIBRARY_PATH","").split(":") + \
            os.environ.get("LD_LIBRARY_PATH","").split(":")
        ,
    )
else:
    print("Compiling for CPU")
    bnm_ext = Extension(
        "cubnm.core",
        [os.path.join("src", "ext", "core.cpp")],
        language="c++",
        extra_compile_args=cpu_extra_compile_args,
        libraries=libraries,
        include_dirs=shared_includes,
        library_dirs=[
            "/opt/miniforge/lib" # added for conda-based cibuildwheel
        ] + \
        os.environ.get("LIBRARY_PATH","").split(":") + \
        os.environ.get("LD_LIBRARY_PATH","").split(":"),
    )

# extend build_ext to also build GSL (if needed) and compile GPU code
class build_ext_gsl_cuda(build_ext):
    def build_extensions(self):
        # Build GSL (if needed)
        # search for libgsl.a and libgslcblas.a in some common paths
        lib_dirs = os.environ.get("LIBRARY_PATH","").split(":") + \
                os.environ.get("LD_LIBRARY_PATH","").split(":") + \
                [
                    "/usr/lib",
                    "/lib",
                    "/usr/local/lib",
                    "/opt/miniforge/lib", # cibuildwheel
                    # TODO: identify and search current conda env
                    os.path.join(os.environ.get('HOME', '/opt'), '.cubnm', 'gsl', 'build', 'lib'), # has been installed before by cuBNM
                ]
        found_gsl = False
        for lib_dir in lib_dirs:
            if ((lib_dir!='') & os.path.exists(lib_dir)):
                r = find_file('libgsl.a', lib_dir)
                if r: # assuming libgsl.a and libgslcblas.a are in the same directory
                    found_gsl = True
                    GSL_LIB_DIR = os.path.dirname(r[0])
                    print(f"Found libgsl.a and libgslcblas.a in {GSL_LIB_DIR}")
                    break
        if not found_gsl:
            print("Downloading and building GSL")
            try:
                gsl_dir = os.path.join(os.environ.get('HOME', '/opt'), '.cubnm', 'gsl')
                os.makedirs(gsl_dir, exist_ok=True)
            except OSError:
                gsl_dir = os.path.join(os.path.abspath(self.build_lib), 'gsl')
                os.makedirs(gsl_dir, exist_ok=True)
            gsl_tar = os.path.join(gsl_dir, 'gsl-2.7.tar.gz')
            gsl_src = os.path.join(gsl_dir, 'gsl-2.7')
            gsl_build = os.path.join(gsl_dir, 'build')
            os.makedirs(gsl_build, exist_ok=True)
            # use gcc (or other default C compilers) as with g++ set as CC
            # GSL compilation fails. Note that this will not affect the compiler
            # of bnm extension as it has already been set to g++
            os.environ["CC"] = _CC
            gsl_setup_commands = [
                f"curl https://mirror.ibcp.fr/pub/gnu/gsl/gsl-2.7.tar.gz -o {gsl_tar}",
                f"cd {gsl_dir} && tar -xf {gsl_tar} &&"
                f"cd {gsl_src} && ./configure --prefix={gsl_build} --enable-shared &&"
                f"make && make install",
            ]
            for command in gsl_setup_commands:
                result = subprocess.run(command, shell=True, stdout=sys.stdout, stderr=subprocess.STDOUT)
                # print(f"Command: {command}")
                # print(f"Return Code: {result.returncode}")
                # print(f"stderr:\n{result.stderr}")
            GSL_LIB_DIR = os.path.join(gsl_build, 'lib')
        # Compile CUDA code into libbnm.a
        if gpu_enabled:
            cuda_dir = os.path.join(ROOT, 'src', 'ext')
            if gpu_model == 'rocm':
                include_flags = " ".join([f"-I {p}" for p in all_includes])
            else:
                include_flags = " ".join([f"-I {p}" for p in shared_includes])
            source_files = ["bnm.cu", "utils.cu", "fc.cu"] + [
                os.path.relpath(f, start=cuda_dir) for f in find_file('*.cu', os.path.join(cuda_dir, 'models'))
            ]
            # create a unified source file including all .cu files
            # this offers significantly better performance than 
            # compiling each file separately and linking them later
            unified_source_path = os.path.join(cuda_dir, "_bnm.cu")
            with open(unified_source_path, 'w') as unified_file:
                for source_file in source_files:
                    unified_file.write(f'#include "{source_file}"\n')
            if gpu_model == 'nvidia':
                compile_commands = [
                    f"nvcc -c -rdc=true --compiler-options '-fPIC' -o {cuda_dir}/_bnm.o {unified_source_path} "
                    f"{include_flags} {' '.join(gpu_extra_compile_args)}",
                    f"nvcc -dlink --compiler-options '-fPIC' -o {cuda_dir}/_bnm_linked.o {cuda_dir}/_bnm.o "
                        f"-L {GSL_LIB_DIR} -lm -lgsl -lgslcblas -lcudart_static",
                    f"ar cru {cuda_dir}/libbnm.a {cuda_dir}/_bnm_linked.o {cuda_dir}/_bnm.o",
                    f"ranlib {cuda_dir}/libbnm.a",
                ]
            else:
                compile_commands = [
                    f"hipcc -fno-gpu-rdc -fPIC {include_flags} {' '.join(gpu_extra_compile_args)} -c {unified_source_path} -o {cuda_dir}/_bnm.o ",
                    f"hipcc --shared -fPIC -fno-gpu-rdc -o {cuda_dir}/_bnm_linked.so {cuda_dir}/_bnm.o "
                        f"-L {GSL_LIB_DIR} -lgsl -lgslcblas -L {ROCM_PATH}/lib -lamdhip64", # -lm and -lcudart_static are removed
                    f"ar cru {cuda_dir}/libbnm.a {cuda_dir}/_bnm_linked.so {cuda_dir}/_bnm.o",
                    f"ranlib {cuda_dir}/libbnm.a",
                ]
            for command in compile_commands:
                print(command)
                result = subprocess.run(command, shell=True, stdout=sys.stdout, stderr=subprocess.STDOUT)
                print(f"Return Code: {result.returncode}")
                if result.stderr:
                    print(f"Standard Error:\n{result.stderr}")
        # Continue with Python extension build
        # add libgsl.a and libgslcblas.a to the compiler objects for
        # their explicit linking
        self.compiler.objects.append(os.path.join(GSL_LIB_DIR, 'libgsl.a'))
        self.compiler.objects.append(os.path.join(GSL_LIB_DIR, 'libgslcblas.a'))
        # self.compiler.add_library_dir(GSL_LIB_DIR) # uncomment for dynamic linking of gsl
        super().build_extensions()

setup(
    version=versioneer.get_version(),
    ext_modules=[bnm_ext],
    cmdclass=versioneer.get_cmdclass({
        'build_ext': build_ext_gsl_cuda,
    }),
)

# restore OS's original $CC and $CXX
os.environ['CC'] = _CC
os.environ['CXX'] = _CXX


"""
Command line interface for cuBNM
"""
import os
import pandas as pd
from cubnm import datasets, optimize, cli_parser

def main():
    parser = cli_parser.get_parser()
    # parse the arguments
    args = parser.parse_args()

    if args.cmd not in ['grid', 'optimize']:
        parser.print_help()
        return
    
    # additional validations
    # make sure either bold, fc or fcd are provided
    if (args.emp_fc_tril is None) and (args.emp_fcd_tril is None) and (args.emp_bold is None):
        parser.error("At least one of the following must be provided: --emp_fc_tril, --emp_fcd_tril, --emp_bold")

    if not args.no_print_args:
        # pretty print the command and its arguments
        if args.cmd == 'grid':
            print("Running grid search with the following options:")
        if args.cmd == 'optimize':
            print("Running evolutionary optimization with the following options:")
        args_df = (
            pd.Series(vars(args))
            .fillna('-')
            .rename('Value').to_frame()
            .drop(index='cmd')
            .rename_axis(index='Option')
        )
        print(args_df.to_markdown(tablefmt="simple_outline"))
    del args.no_print_args
    
    # additional refinement of arguments
    args.do_fc = not args.no_fc
    args.do_fcd = not args.no_fcd
    args.fcd_drop_edges = not args.fcd_keep_edges
    del args.no_fc, args.no_fcd, args.fcd_keep_edges

    # add +/- to gof terms
    for i, term in enumerate(args.gof_terms):
        if term == 'fc_corr':
            args.gof_terms[i] = '+' + term
        else:
            args.gof_terms[i] = '-' + term

    # use example input data if requested
    if args.sc == 'example':
        args.sc = datasets.load_sc('strength', 'schaefer-100')
    if args.sc_dist == 'example':
        args.sc_dist = datasets.load_sc('length', 'schaefer-100')
    if args.emp_bold == 'example':
        args.emp_bold = datasets.load_bold('schaefer-100')
    if (args.cmd == 'optimize') and (args.maps == 'example'):
        args.maps = datasets.load_maps(
            ['myelinmap', 'fcgradient01'],
            'schaefer-100'
        )

    if args.cmd == 'grid':
        run_grid(args)
    elif args.cmd == 'optimize':
        # additional refinement of specific arguments
        if args.het_params is None:
            args.het_params = []
        run_optimize(args)

def _get_problem(args):
    """
    Helper function to get the optimization problem based on CLI inputs

    Parameters
    ----------
    args : :obj:`argparse.NameSpace`

    Returns
    -------
    :obj:`optimize.BNMProblem`
    """
    problem = optimize.BNMProblem(
        model=args.model,
        params=args.params,
        emp_fc_tril=args.emp_fc_tril,
        emp_fcd_tril=args.emp_fcd_tril,
        emp_bold=args.emp_bold,
        do_fc=args.do_fc,
        do_fcd=args.do_fcd,
        het_params=args.het_params,
        maps=args.maps,
        maps_coef_range=args.maps_coef_range,
        node_grouping=args.node_grouping,
        multiobj=args.multiobj,
        sc=args.sc,
        sc_dist=args.sc_dist,
        out_dir=args.out_dir,
        duration=args.duration,
        TR=args.TR,
        dt=args.dt,
        bw_dt=args.bw_dt,
        states_ts=args.states_ts,
        states_sampling=args.states_sampling,
        window_size=args.window_size,
        window_step=args.window_step,
        sim_seed=args.sim_seed,
        exc_interhemispheric=args.exc_interhemispheric,
        force_gpu=args.force_gpu,
        force_cpu=args.force_cpu,
        gof_terms=args.gof_terms,
        bw_params=args.bw_params,
        bold_remove_s=args.bold_remove_s,
        fcd_drop_edges=args.fcd_drop_edges,
        noise_segment_length=args.noise_segment_length,
        sim_verbose=args.sim_verbose,
        progress_interval=args.progress_interval,
    )
    return problem

def run_grid(args):
    """
    Runs a grid search based on CLI inputs

    Parameters
    ----------
    args : :obj:`argparse.NameSpace`
    """
    # initialize optimization problem
    problem = _get_problem(args)
    # initialize and run grid search
    grid = optimize.GridOptimizer()
    grid.optimize(problem, args.grid_shape)
    grid.save()
    print("Optimal simulation:", grid.opt, sep="\n")

def run_optimize(args):
    """
    Runs an evolutionary optimization based on CLI inputs

    Parameters
    ----------
    args : :obj:`argparse.NameSpace`
    """
    # initialize optimization problem
    problem = _get_problem(args)
    # initialize optimizer
    optimizer_cls = getattr(optimize, f'{args.optimizer}Optimizer')
    optimizer = optimizer_cls(
        popsize=args.popsize,
        n_iter=args.n_iter,
        seed=args.optimizer_seed
    )
    # register problem with optimizer
    optimizer.setup_problem(problem)
    # run optimizer and save it
    optimizer.optimize()
    optimizer.save()
    print("Optimal simulation:", optimizer.opt, sep="\n")


if __name__ == '__main__':
    main()

